# mypy: disable-error-code="operator"
from __future__ import annotations

import dataclasses
import functools
import itertools
import math
from typing import NamedTuple, cast

import more_itertools
import portion

from libresvip.core.tick_counter import shift_tempo_list
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import ParamCurve, Points, SongTempo
from libresvip.model.point import Point
from libresvip.utils.music_math import hz2midi, midi2hz
from libresvip.utils.search import find_last_index
from libresvip.utils.translation import gettext_lazy as _

from .constants import (
    MIN_DATA_LENGTH,
    TEMP_VALUE_AS_NULL,
    TIME_UNIT_AS_TICKS_PER_BPM,
)


class VoiSonaParamEvent(NamedTuple):
    idx: int | None
    repeat: int | None
    value: float


class VoiSonaParamEventFloat(NamedTuple):
    idx: float | None
    repeat: float | None
    value: float | None

    @classmethod
    def from_event(cls, event: VoiSonaParamEvent) -> VoiSonaParamEventFloat:
        return cls(
            float(event.idx) if event.idx is not None else None,
            float(event.repeat) if event.repeat is not None else None,
            event.value,
        )


@dataclasses.dataclass
class VoiSonaTrackPitchData:
    events: list[VoiSonaParamEvent]
    tempos: list[SongTempo]
    tick_prefix: int
    vibrato_amplitude_events: list[VoiSonaParamEvent] = dataclasses.field(default_factory=list)
    vibrato_frequency_events: list[VoiSonaParamEvent] = dataclasses.field(default_factory=list)

    @property
    def length(self) -> int:
        last_has_index = find_last_index(self.events, lambda event: event.idx is not None)
        length = self.events[last_has_index].idx + sum(
            event.repeat or 1 for event in self.events[last_has_index:]
        )
        return length + MIN_DATA_LENGTH


def pitch_from_voisona_track(
    data: VoiSonaTrackPitchData,
) -> ParamCurve | None:
    converted_points = [Point.start_point()]
    current_value = -100

    synchronizer = TimeSynchronizer(data.tempos)
    vibrato_amplitude_interval_dict = build_voisona_param_interval_dict(
        data.vibrato_amplitude_events, synchronizer, data.tick_prefix
    )
    vibrato_value_interval_dict = build_voisona_wave_interval_dict(
        data.vibrato_frequency_events, synchronizer, data.tick_prefix
    )

    events_normalized = shape_events(
        normalize_to_tick(append_ending_points(data.events), data.tempos, data.tick_prefix)
    )

    next_pos = None
    for event in events_normalized:
        pos = int(cast("float", event.idx)) - data.tick_prefix
        secs = synchronizer.get_actual_secs_from_ticks(pos)
        length = event.repeat
        try:
            value = round(hz2midi(math.e**event.value) * 100) if event.value is not None else -100
            if value != current_value or next_pos != pos:
                converted_points.append(Point(x=round(pos), y=value))
                if value == -100:
                    converted_points.append(Point(x=round(pos), y=value))
                current_value = value
            secs_step = synchronizer.get_duration_secs_from_ticks(pos, pos + 5)
            for pos_x in range(pos, int(cast("float", pos + length)), 5):
                if value_diff := vibrato_value_interval_dict.get(secs):
                    value_diff *= vibrato_amplitude_interval_dict.get(secs, 1)
                    converted_points.append(Point(x=pos_x, y=round(value + value_diff)))
                else:
                    break
                secs += secs_step
        except OverflowError:
            show_warning(_("Pitch value is out of bounds"))
        next_pos = pos + length
    converted_points.append(Point.end_point())

    return ParamCurve(points=Points(root=converted_points)) if len(converted_points) > 2 else None


def append_ending_points(
    events: list[VoiSonaParamEvent],
) -> list[VoiSonaParamEvent]:
    result = []
    next_pos = None
    for event in events:
        pos = event.idx if event.idx is not None else next_pos
        if pos is None:
            continue
        length = event.repeat if event.repeat is not None else 1
        if next_pos is not None and next_pos < pos:
            result.append(VoiSonaParamEvent(next_pos, None, TEMP_VALUE_AS_NULL))
        result.append(VoiSonaParamEvent(pos, length, event.value))
        next_pos = pos + length
    if next_pos is not None:
        result.append(VoiSonaParamEvent(next_pos, None, TEMP_VALUE_AS_NULL))
    return result


def normalize_to_tick(
    events: list[VoiSonaParamEvent],
    tempo_list: list[SongTempo],
    tick_prefix: int,
) -> list[VoiSonaParamEventFloat]:
    tempos = expand(tempo_list, tick_prefix)
    events_normalized: list[VoiSonaParamEventFloat] = []
    current_tempo_index = 0
    next_pos = 0.0
    next_tick_pos = 0.0
    for _event in events:
        event = VoiSonaParamEventFloat.from_event(_event)
        pos = event.idx if event.idx is not None else next_pos
        if event.idx is None:
            tick_pos = next_tick_pos
        else:
            while (
                current_tempo_index + 1 < len(tempos)
                and tempos[current_tempo_index + 1][0] <= event.idx
            ):
                current_tempo_index += 1
            ticks_in_time_unit = TIME_UNIT_AS_TICKS_PER_BPM * tempos[current_tempo_index][2]
            tick_pos = (
                tempos[current_tempo_index][1]
                + (event.idx - tempos[current_tempo_index][0]) * ticks_in_time_unit
            )
        repeat = event.repeat if event.repeat is not None else 1.0
        remaining_repeat = repeat
        repeat_in_ticks = 0.0
        while (
            current_tempo_index + 1 < len(tempos)
            and tempos[current_tempo_index + 1][0] < pos + repeat
        ):
            repeat_in_ticks += tempos[current_tempo_index + 1][1] - max(
                tempos[current_tempo_index][1], tick_pos
            )
            remaining_repeat -= tempos[current_tempo_index + 1][0] - max(
                tempos[current_tempo_index][0], pos
            )
            current_tempo_index += 1
        repeat_in_ticks += (
            remaining_repeat * TIME_UNIT_AS_TICKS_PER_BPM * tempos[current_tempo_index][2]
        )
        next_pos = pos + repeat
        next_tick_pos = tick_pos + repeat_in_ticks
        events_normalized.append(VoiSonaParamEventFloat(tick_pos, repeat_in_ticks, event.value))
    return [
        VoiSonaParamEventFloat(
            tick.idx + tick_prefix,
            tick.repeat,
            tick.value if tick.value != TEMP_VALUE_AS_NULL else None,
        )
        for tick in events_normalized
    ]


def shape_events(
    events_with_full_params: list[VoiSonaParamEventFloat],
) -> list[VoiSonaParamEventFloat]:
    result: list[VoiSonaParamEventFloat] = []
    for event in events_with_full_params:
        if event.repeat is not None and event.repeat > 0:
            if result:
                last = result[-1]
                if last.idx == event.idx:
                    result[-1] = event
                else:
                    result.append(event)
            else:
                result.append(event)
    return result


def expand(tempos: list[SongTempo], tick_prefix: int) -> list[tuple[int, int, float]]:
    result: list[tuple[int, int, float]] = []
    for i, tempo in enumerate(tempos):
        if i == 0:
            result.append((0, tick_prefix, tempo.bpm))
        else:
            last_pos, last_tick_pos, last_bpm = result[-1]
            ticks_in_time_unit = TIME_UNIT_AS_TICKS_PER_BPM * last_bpm
            new_pos = last_pos + (tempo.position - last_tick_pos) / ticks_in_time_unit
            result.append((int(new_pos), tempo.position, tempo.bpm))
    return result


def vibrato_curve(value: float, shift: float, omega: float, phase: float) -> float:
    return math.sin(omega * (value - shift) + phase)


def build_voisona_param_interval_dict(
    events: list[VoiSonaParamEvent],
    synchronizer: TimeSynchronizer,
    tick_prefix: int,
) -> PiecewiseIntervalDict:
    param_interval_dict = PiecewiseIntervalDict()
    for continuous_part in more_itertools.split_before(
        events,
        lambda event: event.idx is not None,
    ):
        for prev_event, next_event in more_itertools.windowed(
            more_itertools.prepend(
                None,
                normalize_to_tick(
                    append_ending_points(continuous_part),
                    synchronizer.tempo_list,
                    tick_prefix,
                ),
            ),
            2,
        ):
            if next_event is None:
                continue
            next_start = synchronizer.get_actual_secs_from_ticks(next_event.idx - tick_prefix)
            if (
                prev_event is not None
                and (
                    prev_end := synchronizer.get_actual_secs_from_ticks(
                        prev_event.idx + (prev_event.repeat or 1) - tick_prefix
                    )
                )
                and prev_end < next_start
            ):
                param_interval_dict[portion.closedopen(prev_end, next_start)] = next_event.value
            param_interval_dict[
                portion.closedopen(
                    next_start,
                    synchronizer.get_actual_secs_from_ticks(
                        next_event.idx + (next_event.repeat or 1) - tick_prefix
                    ),
                )
            ] = next_event.value
    return param_interval_dict


def build_voisona_wave_interval_dict(
    events: list[VoiSonaParamEvent],
    synchronizer: TimeSynchronizer,
    tick_prefix: int,
) -> PiecewiseIntervalDict:
    param_interval_dict = PiecewiseIntervalDict()
    omega = math.tau * 6
    for continuous_part in more_itertools.split_before(
        events,
        lambda event: event.idx is not None,
    ):
        phase = 0.0
        for prev_event, next_event in more_itertools.windowed(
            more_itertools.prepend(
                None,
                normalize_to_tick(
                    append_ending_points(continuous_part),
                    synchronizer.tempo_list,
                    tick_prefix,
                ),
            ),
            2,
        ):
            next_start = synchronizer.get_actual_secs_from_ticks(next_event.idx - tick_prefix)
            next_end = synchronizer.get_actual_secs_from_ticks(
                next_event.idx + (next_event.repeat or 1) - tick_prefix
            )
            if (
                prev_event is not None
                and (
                    prev_end := synchronizer.get_actual_secs_from_ticks(
                        prev_event.idx + (prev_event.repeat or 1) - tick_prefix
                    )
                )
                and prev_end < next_start
            ):
                prev_start = synchronizer.get_actual_secs_from_ticks(prev_event.idx - tick_prefix)
                param_interval_dict[portion.closedopen(prev_end, next_start)] = vibrato_curve(
                    prev_end, prev_start, omega, phase
                )
            omega = math.tau * (next_event.value or 6)
            param_interval_dict[portion.closedopen(next_start, next_end)] = functools.partial(
                vibrato_curve, shift=next_start, omega=omega, phase=phase
            )
            phase += (next_end - next_start) * omega
    return param_interval_dict


def generate_for_voisona(
    pitch: ParamCurve, tempos: list[SongTempo], tick_prefix: int
) -> VoiSonaTrackPitchData | None:
    events_with_full_params = []
    for i, this_point in enumerate(pitch.points.root):
        next_point = pitch.points[i + 1] if i + 1 < len(pitch.points) else None
        end_tick = next_point.x - tick_prefix if next_point else None
        index = this_point.x - tick_prefix
        repeat = end_tick - index if end_tick else 1
        repeat = max(repeat, 1)
        value = math.log(midi2hz(this_point.y / 100)) if this_point.y != -100 else None
        if value is not None and (next_point is None or next_point.y != -100):
            events_with_full_params.append(
                VoiSonaParamEventFloat(float(index), float(repeat), float(value))
            )
    are_events_connected_to_next = [
        this_event.idx + this_event.repeat >= next_event.idx if next_event else False
        for this_event, next_event in zip(
            events_with_full_params, events_with_full_params[1:] + [None]
        )
    ]
    events = denormalize_from_tick(events_with_full_params, tempos, tick_prefix)
    events = restore_connection(events, are_events_connected_to_next)
    events = merge_events_if_possible(events)
    events = remove_redundant_index(events)
    events = remove_redundant_repeat(events)
    if not events:
        return None
    last_event_with_index = next(
        (event for event in reversed(events) if event.idx is not None), None
    )
    if last_event_with_index is not None:
        length = last_event_with_index.idx
        for event in events[events.index(last_event_with_index) :]:
            length += event.repeat or 1
    return VoiSonaTrackPitchData(events, [], tick_prefix)


def denormalize_from_tick(
    events_with_full_params: list[VoiSonaParamEventFloat],
    tempos_in_ticks: list[SongTempo],
    tick_prefix: int,
) -> list[VoiSonaParamEvent]:
    tempos = expand(
        shift_tempo_list(tempos_in_ticks, tick_prefix),
        tick_prefix,
    )
    events_with_full_params = [
        event if event.idx is None else event._replace(idx=event.idx + tick_prefix)
        for event in events_with_full_params
    ]
    events = []
    current_tempo_index = 0
    tick_pos = None
    for event_double in events_with_full_params:
        if event_double.idx is not None:
            tick_pos = event_double.idx
        if tick_pos is None or event_double.idx is None:
            msg = "Invalid event"
            raise ValueError(msg)
        while (
            current_tempo_index + 1 < len(tempos) and tempos[current_tempo_index + 1][1] < tick_pos
        ):
            current_tempo_index += 1
        ticks_per_time_unit = tempos[current_tempo_index][2] * TIME_UNIT_AS_TICKS_PER_BPM
        pos = (
            tempos[current_tempo_index][0]
            + (event_double.idx - tempos[current_tempo_index][1]) / ticks_per_time_unit
        )
        repeat_in_ticks = event_double.repeat
        repeat = 0.0
        while (current_tempo_index + 1 < len(tempos)) and (
            tempos[current_tempo_index + 1][1] < tick_pos + repeat_in_ticks
        ):
            repeat += tempos[current_tempo_index + 1][0] - max(tempos[current_tempo_index][0], pos)
            repeat_in_ticks -= tempos[current_tempo_index + 1][1] - max(
                tempos[current_tempo_index][1], tick_pos
            )
            current_tempo_index += 1
        repeat += repeat_in_ticks / (TIME_UNIT_AS_TICKS_PER_BPM * tempos[current_tempo_index][2])
        events.append(VoiSonaParamEvent(round(pos), round(max(repeat, 1)), event_double.value or 0))
    return events


def restore_connection(
    events: list[VoiSonaParamEvent], are_events_connected_to_next: list[bool]
) -> list[VoiSonaParamEvent]:
    new_events = []
    for (prev_event, next_event), is_connected_to_next in zip(
        more_itertools.windowed(itertools.chain(events, [None]), 2),
        are_events_connected_to_next,
    ):
        if next_event is None or not is_connected_to_next:
            new_events.append(prev_event)
        else:
            new_events.append(prev_event._replace(repeat=next_event.idx - prev_event.idx))
    return new_events


def merge_events_if_possible(
    events: list[VoiSonaParamEvent],
) -> list[VoiSonaParamEvent]:
    new_events: list[VoiSonaParamEvent] = []
    for event in events:
        if not new_events:
            new_events.append(event)
        else:
            last_event = new_events[-1]
            overlapped_len = last_event.idx + last_event.repeat - event.idx
            if overlapped_len > 0:
                new_events[-1] = new_events[-1]._replace(repeat=event.idx - last_event.idx)
                event = event._replace(
                    idx=overlapped_len + event.idx,
                    repeat=event.repeat - overlapped_len,
                )
                last_event = VoiSonaParamEvent(
                    event.idx, overlapped_len, event.value + last_event.value
                )
                new_events.append(last_event)
            if last_event.value == event.value and last_event.idx + last_event.repeat == event.idx:
                new_events[-1] = new_events[-1]._replace(repeat=last_event.repeat + event.repeat)
            else:
                new_events.append(event)
    return new_events


def remove_redundant_index(
    events: list[VoiSonaParamEvent],
) -> list[VoiSonaParamEvent]:
    new_events: list[VoiSonaParamEvent] = []
    for event in events:
        if not new_events:
            new_events.append(event)
        else:
            prev_event = new_events[-1]
            if (
                prev_event.idx is not None
                and prev_event.repeat is not None
                and prev_event.idx + prev_event.repeat == event.idx
            ):
                new_events.append(event._replace(idx=None))
            else:
                new_events.append(event)
    return new_events


def remove_redundant_repeat(
    events: list[VoiSonaParamEvent],
) -> list[VoiSonaParamEvent]:
    return [event if event.repeat != 1 else event._replace(repeat=None) for event in events]



from typing import Any

from wanakana.japanese import (
    is_kana,
    is_romaji,
    to_hiragana,
    to_katakana,
)
from wanakana.japanese import to_romaji as _to_romaji

__all__ = [
    "is_kana",
    "is_romaji",
    "to_hiragana",
    "to_katakana",
    "to_romaji",
]


def to_romaji(word: str, **kwargs: Any) -> str:
    kwargs.setdefault("custom_romaji_mapping", {"っ": "cl"})
    return _to_romaji(word, **kwargs) if word else ""



import dataclasses
import functools

import more_itertools
import portion

from libresvip.core.constants import DEFAULT_JAPANESE_LYRIC
from libresvip.core.lyric_phoneme.japanese import is_kana
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import Note, ParamCurve, Project, SingingTrack, SongTempo, TimeSignature
from libresvip.utils.music_math import linear_interpolation, midi2hz
from libresvip.utils.text import uuid_str
from libresvip.utils.translation import gettext_lazy as _

from .model import (
    VoiceVoxNote,
    VoiceVoxProject,
    VoiceVoxSong,
    VoiceVoxTempo,
    VoiceVoxTimeSignature,
    VoiceVoxTrack,
)
from .options import OutputOptions


@dataclasses.dataclass
class VOICEVOXGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> VoiceVoxProject:
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        song = VoiceVoxSong(
            tempos=self.generate_tempos(project.song_tempo_list),
            time_signatures=self.generate_time_signatures(project.time_signature_list),
        )
        for track in project.track_list:
            if isinstance(track, SingingTrack):
                track_key = uuid_str()
                song.track_order.append(track_key)
                song.tracks[track_key] = self.generate_track(track)
        return VoiceVoxProject(song=song)

    def generate_time_signatures(
        self, time_signature_list: list[TimeSignature]
    ) -> list[VoiceVoxTimeSignature]:
        return [
            VoiceVoxTimeSignature(
                measure_number=time_signature.bar_index + 1,
                beats=time_signature.numerator,
                beat_type=time_signature.denominator,
            )
            for time_signature in time_signature_list
        ]

    def generate_tempos(self, tempo_list: list[SongTempo]) -> list[VoiceVoxTempo]:
        return [
            VoiceVoxTempo(bpm=round(tempo.bpm), position=tempo.position) for tempo in tempo_list
        ]

    def generate_track(self, track: SingingTrack) -> VoiceVoxTrack:
        return VoiceVoxTrack(
            name=track.title,
            solo=track.solo,
            mute=track.mute,
            gain=track.volume,
            notes=[self.generate_note(note) for note in track.note_list],
            pitch_edit_data=self.generate_pitch(track.edited_params.pitch),
        )

    def generate_note(self, note: Note) -> VoiceVoxNote:
        if not is_kana(note.lyric):
            lyric = DEFAULT_JAPANESE_LYRIC
            msg_prefix = _("Unsupported lyric: ")
            show_warning(f"{msg_prefix} {note.lyric}")
        else:
            lyric = note.lyric
        return VoiceVoxNote(
            lyric=lyric,
            position=note.start_pos,
            duration=note.length,
            note_number=note.key_number,
        )

    def generate_pitch(self, pitch: ParamCurve) -> list[float]:
        frequency_interval_dict = PiecewiseIntervalDict()
        max_pitch_secs = 0.0
        secs_step = 4 / 375
        prev_secs = None
        prev_freq: float = -1
        for point in pitch.points.root:
            if point.y == -100:
                if prev_secs is not None:
                    max_pitch_secs = prev_secs
                prev_secs = None
                prev_freq = -1
            else:
                secs = self.time_synchronizer.get_actual_secs_from_ticks(
                    point.x - self.first_bar_length
                )
                freq = midi2hz(point.y / 100)
                if prev_secs is not None:
                    frequency_interval_dict[portion.openclosed(prev_secs, secs)] = (
                        functools.partial(
                            linear_interpolation,  # type: ignore[call-arg]
                            start=(prev_secs, prev_freq),
                            end=(secs, freq),
                        )
                    )
                else:
                    frequency_interval_dict[portion.singleton(secs)] = freq
                prev_secs = secs
                prev_freq = freq
        return [
            frequency_interval_dict.get(secs, -1)
            for secs in more_itertools.numeric_range(0, max_pitch_secs, secs_step)
        ]



from dataclasses import dataclass, field

from .ds_file import DsItem
from .ds_note import DsNote
from .ds_param_curve import DsParamCurve


@dataclass
class DsProjectModel:
    note_list: list[DsNote] = field(default_factory=list)
    pitch_param_curve: DsParamCurve = field(default_factory=DsParamCurve)
    gender_param_curve: DsParamCurve = field(default_factory=DsParamCurve)

    def to_param_model(self) -> DsItem:
        ds_notes = self.note_list
        input_text = ""
        phoneme_seq = ""
        phoneme_counts = []
        input_note_seq = ""
        input_duration = ""
        input_slur = ""
        is_slur_seq = ""
        phoneme_dur_seq = ""
        phoneme_dur_enabled = True
        for i in range(len(ds_notes)):
            cur_note = ds_notes[i]
            ds_phoneme = cur_note.ds_phoneme
            consonant = ds_phoneme.consonant
            vowel = ds_phoneme.vowel
            input_text += cur_note.lyric.replace("-", "")
            phoneme_counts.append(1)
            if consonant is not None:
                phoneme_seq += f"{consonant.phoneme} "
                if consonant.duration is not None:
                    phoneme_dur_seq += f"{consonant.duration} "
                elif phoneme_dur_enabled:
                    phoneme_dur_enabled = False
                is_slur_seq += "0 "
                phoneme_counts[-1] += 1
            phoneme_seq += vowel.phoneme
            phoneme_dur_seq += str(vowel.duration)
            input_note_seq += vowel.note_name
            input_duration += str(cur_note.duration)
            is_slur_seq += "1" if cur_note.is_slur else "0"
            input_slur += "1" if cur_note.is_slur else "0"
            if i < len(ds_notes) - 1:
                if not cur_note.is_slur:
                    input_text += " "
                input_note_seq += " "
                input_duration += " "
                input_slur += " "
                is_slur_seq += " "
                phoneme_seq += " "
                phoneme_dur_seq += " "

        pitch_points = self.pitch_param_curve.point_list
        f0_sequence = None
        if pitch_points and len(pitch_points) > 0:
            f0_sequence = " ".join(f"{p.value:.1f}" for p in pitch_points)

        gender_points = self.gender_param_curve.point_list
        gender_sequence = None
        if gender_points and len(gender_points) > 0:
            gender_sequence = " ".join(f"{p.value:.1f}" for p in gender_points)

        return DsItem(
            text=input_text,
            ph_seq=phoneme_seq,
            note_seq=input_note_seq,
            note_dur=input_duration,
            note_dur_seq=phoneme_dur_seq,
            note_slur=input_slur,
            is_slur_seq=is_slur_seq,
            ph_dur=phoneme_dur_seq if phoneme_dur_enabled else None,
            ph_num=" ".join(str(phoneme_count) for phoneme_count in phoneme_counts),
            f0_timestep=str(self.pitch_param_curve.step_size),
            f0_seq=f0_sequence,
            gender_timestep=str(self.gender_param_curve.step_size),
            gender=gender_sequence,
            offset=0,
            input_type="phoneme",
        )



import functools
from collections.abc import MutableMapping
from typing import Any

from libresvip.core.config import LyricsReplacement, get_ui_settings
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project, SingingTrack

from .options import ProcessOptions


def replace_lyric(replacement: LyricsReplacement | MutableMapping[str, Any], text: str) -> str:
    if isinstance(replacement, MutableMapping):
        replacement = LyricsReplacement(**replacement)
    return replacement.replace(text)


class RemoveShortSilencesMiddleware(plugin_base.MiddlewareBase):
    def process(self, project: Project, options: ProcessOptions) -> Project:
        settings = get_ui_settings()
        if options.lyric_replacement_preset_name in settings.lyric_replace_rules:
            for track in project.track_list:
                if isinstance(track, SingingTrack):
                    for note in track.note_list:
                        note.lyric = functools.reduce(
                            lambda lyric, replacement: replace_lyric(replacement, lyric),
                            settings.lyric_replace_rules[options.lyric_replacement_preset_name],
                            note.lyric,
                        )
        return project



# Ported from QSvip by SineStriker
import abc
import dataclasses
import enum
import inspect
import math
import struct
from itertools import chain
from typing import Generic, Literal, NamedTuple, TypeVar

from more_itertools import batched

XSItem = TypeVar("XSItem")
MIN_NOTE_DURATION = 0.045
MAX_NOTE_DURATION = 20.0


def to_backing_field(key: str) -> str:
    return f"<{key}>k__BackingField"


class XSLineParamNode(NamedTuple):
    pos: int = 0
    value: int = 0


@dataclasses.dataclass
class XSLineParam:
    """SingingTool.Model.Line.LineParam"""

    line_param: bytes = dataclasses.field(
        default=b"",
        metadata={
            "alias": "LineParam",
        },
    )
    nodes: list[XSLineParamNode] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.nodes = []
        if len(self.line_param) >= 4:
            (node_count,) = struct.unpack("<i", bytearray(self.line_param[:4]))
            self.nodes.extend(
                XSLineParamNode._make(each)
                for each in batched(
                    struct.unpack(
                        f"<{node_count * 2}i",
                        bytearray(self.line_param[4 : 4 + node_count * 8]),
                    ),
                    2,
                )
            )

    def convert_to_param(self) -> None:
        self.line_param = struct.pack("<i", len(self.nodes))
        self.line_param += struct.pack(f"<{len(self.nodes) * 2}i", *chain.from_iterable(self.nodes))
        expected_len = max(64, 2 ** math.ceil(math.log2(len(self.line_param))))
        if len(self.line_param) < expected_len:
            self.line_param += b"\x00" * (expected_len - len(self.line_param))


@dataclasses.dataclass
class XSVibratoStyle:
    """SingingTool.Model.VibratoStyle"""

    amp_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={"alias": "_ampLine", "order": 0},
    )
    freq_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={"alias": "_freqLine", "order": 1},
    )
    is_anti_phase: bool = dataclasses.field(
        default=False,
        metadata={
            "alias": to_backing_field("IsAntiPhase"),
        },
    )


@dataclasses.dataclass
class XSVibratoPercentInfo:
    """SingingTool.Model.VibratoPercentInfo"""

    start_percent: float = dataclasses.field(
        default=0,
        metadata={"alias": "_startPercent", "order": 0},
    )
    end_percent: float = dataclasses.field(
        default=100,
        metadata={"alias": "_endPercent", "order": 1},
    )


class XSReverbPresetEnum(enum.IntEnum):
    NONE = -1
    DEFAULT = enum.auto()
    SMALLHALL1 = enum.auto()
    SMALLHALL2 = enum.auto()
    MEDIUMHALL1 = enum.auto()
    MEDIUMHALL2 = enum.auto()
    LARGEHALL1 = enum.auto()
    LARGEHALL2 = enum.auto()
    SMALLROOM1 = enum.auto()
    SMALLROOM2 = enum.auto()
    MEDIUMROOM1 = enum.auto()
    MEDIUMROOM2 = enum.auto()
    LARGEROOM1 = enum.auto()
    LARGEROOM2 = enum.auto()
    MEDIUMER1 = enum.auto()
    MEDIUMER2 = enum.auto()
    PLATEHIGH = enum.auto()
    PLATELOW = enum.auto()
    LONGREVERB1 = enum.auto()
    LONGREVERB2 = enum.auto()


@dataclasses.dataclass
class XSReverbPreset:
    """SingingTool.Library.Audio.ReverbPreset"""

    value: XSReverbPresetEnum = dataclasses.field(
        default=XSReverbPresetEnum.NONE,
        metadata={
            "alias": "value__",
        },
    )


@dataclasses.dataclass
class XSIOverlappable:
    overlapped: bool = dataclasses.field(
        default=False,
        metadata={"alias": to_backing_field("Overlaped"), "order": 6},
    )


@dataclasses.dataclass
class XSBeatSize:
    """SingingTool.Model.SingingGeneralConcept.BeatSize"""

    x: int = dataclasses.field(
        default=0,
        metadata={"alias": "_x", "order": 0},
    )
    y: int = dataclasses.field(
        default=0,
        metadata={"alias": "_y", "order": 1},
    )


@dataclasses.dataclass
class XSSongBeat(XSIOverlappable):
    """SingingTool.Model.SingingGeneralConcept.SongBeat"""

    bar_index: int = dataclasses.field(
        default=0,
        metadata={
            "alias": "_barIndex",
            "order": 0,
        },
    )
    beat_size: XSBeatSize = dataclasses.field(
        default_factory=XSBeatSize,
        metadata={
            "alias": "_beatSize",
            "order": 1,
        },
    )


@dataclasses.dataclass
class XSSongTempo(XSIOverlappable):
    """SingingTool.Model.SingingGeneralConcept.SongTempo"""

    pos: int = dataclasses.field(
        default=0,
        metadata={
            "alias": "_pos",
            "order": 0,
        },
    )
    tempo: int = dataclasses.field(
        default=120,
        metadata={
            "alias": "_tempo",
            "order": 1,
        },
    )


class XSNoteHeadTagEnum(enum.IntEnum):
    NoTag = 0
    SilTag = 1
    SpTag = 2


@dataclasses.dataclass
class XSNoteHeadTag:
    """SingingTool.Model.NoteHeadTag"""

    value: XSNoteHeadTagEnum = dataclasses.field(
        default=XSNoteHeadTagEnum.NoTag,
        metadata={
            "alias": "value__",
        },
    )


@dataclasses.dataclass
class XSNotePhoneInfo:
    """SingingTool.Model.NotePhoneInfo"""

    head_phone_time_in_sec: float = dataclasses.field(
        default=0,
        metadata={
            "alias": to_backing_field("HeadPhoneTimeInSec"),
            "order": 0,
        },
    )
    mid_part_over_tail_part_ratio: float = dataclasses.field(
        default=0,
        metadata={
            "alias": to_backing_field("MidPartOverTailPartRatio"),
            "order": 1,
        },
    )


@dataclasses.dataclass
class XSNote(XSIOverlappable):
    """SingingTool.Model.Note"""

    start_pos: int = dataclasses.field(
        default=0,
        metadata={"alias": "_startPos", "order": 0},
    )
    width_pos: int = dataclasses.field(
        default=480,
        metadata={"alias": "_widthPos", "order": 1},
    )
    key_index: int = dataclasses.field(
        default=60,
        metadata={"alias": "_keyIndex", "order": 2},
    )
    lyric: str = dataclasses.field(
        default="",
        metadata={"alias": "_lyric", "order": 3},
    )
    pronouncing: str = dataclasses.field(
        default="",
        metadata={"alias": "_pronouncing", "order": 4},
    )
    head_tag: XSNoteHeadTag = dataclasses.field(
        default_factory=XSNoteHeadTag,
        metadata={"alias": "_headTag", "order": 5},
    )
    note_phone_info: XSNotePhoneInfo | None = dataclasses.field(
        default=None,
        metadata={"alias": to_backing_field("NotePhoneInfo"), "order": 7},
    )
    vibrato_percent: int = dataclasses.field(
        default=0,
        metadata={"alias": to_backing_field("VibratoPercent"), "order": 8},
    )
    vibrato: XSVibratoStyle | None = dataclasses.field(
        default=None,
        metadata={"alias": to_backing_field("Vibrato"), "order": 9},
    )
    vibrato_percent_info: XSVibratoPercentInfo | None = dataclasses.field(
        default=None,
        metadata={
            "alias": to_backing_field("VibratoPercentInfo"),
            "order": 10,
        },
    )


@dataclasses.dataclass
class XSBuf(Generic[XSItem]):
    """System.Collections.Generic.List"""

    items: list[XSItem] = dataclasses.field(
        default_factory=list,
        metadata={
            "alias": "_items",
            "order": 0,
        },
    )
    size: int = dataclasses.field(
        default=0,
        metadata={
            "alias": "_size",
            "order": 1,
        },
    )
    version: int = dataclasses.field(
        default=0,
        metadata={
            "alias": "_version",
            "order": 2,
        },
    )


@dataclasses.dataclass
class XSBufList(Generic[XSItem]):
    """SingingTool.Library.SerialOverlapableItemList"""

    buf: XSBuf[XSItem] = dataclasses.field(
        default_factory=XSBuf[XSItem],
        metadata={
            "alias": "_buf",
            "order": 0,
        },
    )
    buf_1: XSBuf[XSItem] = dataclasses.field(
        default_factory=XSBuf[XSItem],
        metadata={
            "alias": "SerialItemList`1+_buf",
            "order": 1,
        },
    )


class XSTrackType(enum.Enum):
    Singing = enum.auto()
    Instrument = enum.auto()


@dataclasses.dataclass
class XSITrack(abc.ABC):
    """SingingTool.Model.ITrack"""

    track_type: XSTrackType
    pan: float = dataclasses.field(
        default=0,
        metadata={
            "alias": "_pan",
            "order": 9,
        },
    )
    name: str = dataclasses.field(
        default="",
        metadata={
            "alias": "_name",
            "order": 10,
        },
    )
    mute: bool = dataclasses.field(
        default=False,
        metadata={
            "alias": "_mute",
            "order": 11,
        },
    )
    solo: bool = dataclasses.field(
        default=False,
        metadata={
            "alias": "_solo",
            "order": 12,
        },
    )
    volume: float = dataclasses.field(
        default=0.7,
        metadata={
            "alias": "_volume",
            "order": 8,
        },
    )


@dataclasses.dataclass
class XSSingingTrack(XSITrack):  # type: ignore[override]
    """SingingTool.Model.SingingTrack"""

    track_type: Literal[XSTrackType.Singing] = XSTrackType.Singing
    note_list: XSBufList[XSNote] = dataclasses.field(
        default_factory=XSBufList[XSNote],
        metadata={
            "alias": "_noteList",
            "order": 0,
        },
    )
    need_refresh_base_metadata_flag: bool = dataclasses.field(
        default=False,
        metadata={
            "alias": "_needRefreshBaseMetadataFlag",
            "order": 1,
        },
    )
    edited_pitch_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={
            "alias": "_editedPitchLine",
            "order": 2,
        },
    )
    edited_volume_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={
            "alias": "_editedVolumeLine",
            "order": 3,
        },
    )
    edited_breath_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={
            "alias": "_editedBreathLine",
            "order": 4,
        },
    )
    edited_gender_line: XSLineParam = dataclasses.field(
        default_factory=XSLineParam,
        metadata={
            "alias": "_editedGenderLine",
            "order": 5,
        },
    )
    edited_power_line: XSLineParam | None = dataclasses.field(
        default=None,
        metadata={
            "alias": "_editedPowerLine",
            "order": 6,
        },
    )
    reverb_preset: XSReverbPreset = dataclasses.field(
        default_factory=XSReverbPreset,
        metadata={
            "alias": "_reverbPreset",
            "order": 7,
        },
    )
    ai_singer_id: str = dataclasses.field(
        default="",
        metadata={
            "alias": to_backing_field("AISingerId"),
            "order": 13,
        },
    )


@dataclasses.dataclass
class XSInstrumentTrack(XSITrack):  # type: ignore[override]
    """SingingTool.Model.InstrumentTrack"""

    track_type: Literal[XSTrackType.Instrument] = XSTrackType.Instrument
    volume: float = dataclasses.field(
        default=0.3,
        metadata={
            "alias": "_volume",
            "order": 8,
        },
    )
    sample_rate: float = dataclasses.field(
        default=48000,
        metadata={
            "alias": to_backing_field("SampleRate"),
            "order": 14,
        },
    )
    sample_count: int = dataclasses.field(
        default=0,
        metadata={
            "alias": to_backing_field("SampleCount"),
            "order": 15,
        },
    )
    channel_count: int = dataclasses.field(
        default=0,
        metadata={
            "alias": to_backing_field("ChannelCount"),
            "order": 16,
        },
    )
    offset_in_pos: int = dataclasses.field(
        default=0,
        metadata={
            "alias": to_backing_field("OffsetInPos"),
            "order": 17,
        },
    )
    instrument_file_path: str = dataclasses.field(
        default="",
        metadata={
            "alias": to_backing_field("InstrumentFilePath"),
            "order": 18,
        },
    )


@dataclasses.dataclass
class XSAppModel:
    """SingingTool.Model.AppModel"""

    project_file_path: str = dataclasses.field(
        default="",
        metadata={
            "alias": to_backing_field("ProjectFilePath"),
            "order": 0,
        },
    )
    tempo_list: XSBufList[XSSongTempo] = dataclasses.field(
        default_factory=XSBufList[XSSongTempo],
        metadata={
            "alias": "_tempoList",
            "order": 1,
        },
    )
    beat_list: XSBufList[XSSongBeat] = dataclasses.field(
        default_factory=XSBufList[XSSongBeat],
        metadata={
            "alias": "_beatList",
            "order": 2,
        },
    )
    track_list: XSBuf[XSITrack] = dataclasses.field(
        default_factory=XSBuf[XSITrack],
        metadata={
            "alias": "_trackList",
            "order": 3,
        },
    )
    quantize: int = dataclasses.field(
        default=8,
        metadata={
            "alias": "_quantize",
            "order": 4,
        },
    )
    is_triplet: bool = dataclasses.field(
        default=False,
        metadata={
            "alias": "_isTriplet",
            "order": 5,
        },
    )
    is_numerical_key_name: bool = dataclasses.field(
        default=True,
        metadata={
            "alias": "_isNumerialKeyName",
            "order": 6,
        },
    )
    first_numerical_key_name_at_index: int = dataclasses.field(
        default=0,
        metadata={
            "alias": "_firstNumerialKeyNameAtIndex",
            "order": 7,
        },
    )
    actual_project_file_path: str | None = dataclasses.field(
        default=None,
        metadata={
            "alias": to_backing_field("ActualProjectFilePath"),
            "order": 8,
        },
    )


fullname2classes = {
    inspect.getdoc(cls): cls
    for cls in (
        XSAppModel,
        XSBeatSize,
        XSSongBeat,
        XSITrack,
        XSLineParam,
        XSBufList,
        XSSongTempo,
        XSITrack,
        XSBuf,
        XSSingingTrack,
        XSInstrumentTrack,
        XSVibratoPercentInfo,
        XSVibratoStyle,
        XSNotePhoneInfo,
        XSNoteHeadTag,
        XSNote,
        XSReverbPreset,
    )
}



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis

from .model import Y77Project
from .options import InputOptions, OutputOptions
from .y77_generator import Y77Generator
from .y77_parser import Y77Parser


class Y77Converter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        y77_project = Y77Project.model_validate_json(path.read_bytes())
        return Y77Parser(options).parse_project(y77_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        if len(project.song_tempo_list) != 1:
            project = reset_time_axis(project, options.tempo)
        y77_project = Y77Generator(options).generate_project(project)
        path.write_bytes(
            json.dumps(y77_project.model_dump(mode="json", by_alias=True)).encode("utf-8")
        )



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VoiceVoxProject
from .options import InputOptions, OutputOptions
from .voicevox_generator import VOICEVOXGenerator
from .voicevox_parser import VOICEVOXParser


class VOICEVOXConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        voicevox_project = VoiceVoxProject.model_validate_json(path.read_bytes())
        return VOICEVOXParser(options).parse_project(voicevox_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        voicevox_project = VOICEVOXGenerator(options).generate_project(project)
        path.write_bytes(
            json.dumps(voicevox_project.model_dump(mode="json", by_alias=True)).encode("utf-8")
        )



import dataclasses
import operator

import mido_fix as mido

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.text import SYMBOL_PATTERN

from .constants import ControlChange
from .midi_pitch import generate_for_midi
from .options import OutputOptions


@dataclasses.dataclass
class MidiGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    @property
    def tick_rate(self) -> float:
        return TICKS_IN_BEAT / self.options.ticks_per_beat

    def generate_project(self, project: Project) -> mido.MidiFile:
        self.first_bar_length = round(
            project.time_signature_list[0].bar_length(self.options.ticks_per_beat)
        )
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        mido_obj = mido.MidiFile(charset=self.options.lyric_encoding)
        mido_obj.ticks_per_beat = self.options.ticks_per_beat
        master_track = mido.MidiTrack()
        self.generate_tempos(master_track, project.song_tempo_list)
        self.generate_time_signatures(master_track, project.time_signature_list)
        master_track.sort(key=operator.attrgetter("time"))
        mido_obj.tracks.append(master_track)
        mido_obj.tracks.extend(self.generate_tracks(project.track_list))
        self._convert_cumulative_to_delta(mido_obj.tracks)
        return mido_obj

    @staticmethod
    def _convert_cumulative_to_delta(tracks: list[mido.MidiTrack]) -> None:
        for track in tracks:
            tick = 0
            for event in track:
                tick, event.time = event.time, event.time - tick

    def generate_tempos(
        self, master_track: mido.MidiTrack, song_tempo_list: list[SongTempo]
    ) -> None:
        for tempo in song_tempo_list:
            if tempo.position >= 0:
                master_track.append(
                    mido.MetaMessage(
                        "set_tempo",
                        tempo=mido.bpm2tempo(tempo.bpm),
                        time=round(tempo.position / self.tick_rate),
                    )
                )

    def generate_time_signatures(
        self,
        master_track: mido.MidiTrack,
        time_signature_list: list[TimeSignature],
    ) -> None:
        prev_ticks = 0
        prev_time_signature = None
        for time_signature in time_signature_list:
            if time_signature.bar_index >= 0:
                if prev_time_signature is not None:
                    prev_ticks += round(
                        (time_signature.bar_index - prev_time_signature.bar_index)
                        * prev_time_signature.bar_length(self.options.ticks_per_beat)
                    )
                master_track.append(
                    mido.MetaMessage(
                        "time_signature",
                        numerator=time_signature.numerator,
                        denominator=time_signature.denominator,
                        time=prev_ticks,
                    )
                )
                prev_time_signature = time_signature

    def generate_tracks(self, tracks: list[Track]) -> list[mido.MidiTrack]:
        return [
            mido_track
            for track in tracks
            if isinstance(track, SingingTrack)
            and (mido_track := self.generate_track(track)) is not None
        ]

    def generate_track(self, track: SingingTrack) -> mido.MidiTrack:
        lyrics = [
            note.pronunciation if note.pronunciation is not None else note.lyric
            for note in track.note_list
        ]
        if self.options.remove_symbols:
            lyrics = [SYMBOL_PATTERN.sub("", lyric) for lyric in lyrics]
        pinyins = get_pinyin_series(lyrics)
        mido_track = mido.MidiTrack()
        mido_track.name = track.title
        for i, note in enumerate(track.note_list):
            if self.options.export_lyrics:
                mido_track.append(
                    mido.MetaMessage(
                        "lyrics",
                        text=(pinyins[i] if self.options.compatible_lyric else lyrics[i]),
                        time=round(note.start_pos / self.tick_rate),
                    )
                )
            mido_track.append(
                mido.Message(
                    "note_on",
                    note=note.key_number,
                    time=round(note.start_pos / self.tick_rate),
                )
            )
            mido_track.append(
                mido.Message(
                    "note_off",
                    note=note.key_number,
                    time=round(note.end_pos / self.tick_rate),
                )
            )
        if pitch_data := generate_for_midi(
            self.first_bar_length,
            track.edited_params.pitch,
            track.note_list,
            self.synchronizer,
        ):
            for pbs_event in pitch_data.pbs:
                msg_time = round(pbs_event.tick / self.tick_rate)
                mido_track.extend(
                    [
                        mido.Message(
                            "control_change",
                            control=ControlChange.RPN_MSB.value,
                            value=0,
                            time=msg_time,
                        ),
                        mido.Message(
                            "control_change",
                            control=ControlChange.RPN_LSB.value,
                            value=0,
                            time=msg_time,
                        ),
                        mido.Message(
                            "control_change",
                            control=ControlChange.DATA_ENTRY.value,
                            value=pbs_event.value,
                            time=msg_time,
                        ),
                    ]
                )
            for pitch_event in pitch_data.pit:
                mido_track.append(
                    mido.Message(
                        "pitchwheel",
                        pitch=pitch_event.value,
                        time=round(pitch_event.tick / self.tick_rate),
                    )
                )
        mido_track.sort(key=operator.attrgetter("time"))
        if len(mido_track):
            return mido_track



from operator import attrgetter

from libresvip.core.time_interval import RangeInterval
from libresvip.model.base import Note


def has_overlap(notes: list[Note]) -> bool:
    if len(notes) < 2:
        return False
    notes.sort(key=attrgetter("start_pos"))
    interval = RangeInterval()
    for note in notes:
        if interval.includes(note.start_pos):
            return True
        interval |= RangeInterval([(note.start_pos, note.end_pos)])
    return False



from typing import Any

import yaml
from yaml.tokens import (
    AliasToken,
    AnchorToken,
    FlowMappingEndToken,
    FlowMappingStartToken,
    KeyToken,
    ScalarToken,
    ValueToken,
)

from libresvip.utils.yamlutils.common import hash_key, set_yaml_grammar

try:
    from yaml import CSafeLoader as DefaultSafeLoader
except ImportError:
    from yaml import SafeLoader as DefaultSafeLoader


class CustomLoader(yaml.SafeLoader):
    expand_aliases = False

    def emit_yq_kv(self, key: Any, value: Any, original_token: yaml.tokens.Token) -> None:
        marks = {
            "start_mark": original_token.start_mark,
            "end_mark": original_token.end_mark,
        }
        self.tokens.append(FlowMappingStartToken(**marks))
        self.tokens.append(KeyToken(**marks))
        self.tokens.append(ScalarToken(value=key, plain=True, **marks))
        self.tokens.append(ValueToken(**marks))
        self.tokens.append(ScalarToken(value=value, plain=True, **marks))
        self.tokens.append(FlowMappingEndToken(**marks))

    def fetch_alias(self) -> Any:
        if self.expand_aliases:
            return super().fetch_alias()
        self.save_possible_simple_key()
        self.allow_simple_key = False
        alias_token = self.scan_anchor(AliasToken)
        self.emit_yq_kv("__yq_alias__", alias_token.value, original_token=alias_token)

    def fetch_anchor(self) -> Any:
        if self.expand_aliases:
            return super().fetch_anchor()
        self.save_possible_simple_key()
        self.allow_simple_key = False
        anchor_token = self.scan_anchor(AnchorToken)  # noqa: F841
        # self.emit_yq_kv("__yq_anchor__", anchor_token.value, original_token=anchor_token)


def get_loader(
    use_annotations: bool = False,
    expand_aliases: bool = True,
    expand_merge_keys: bool = True,
) -> yaml.SafeLoader:
    def construct_sequence(loader: yaml.SafeLoader, node: yaml.nodes.SequenceNode) -> list[Any]:
        annotations = []
        for i, v_node in enumerate(node.value):
            if not use_annotations:
                break
            if (
                v_node.tag
                and v_node.tag.startswith("!")
                and not v_node.tag.startswith("!!")
                and len(v_node.tag) > 1
            ):
                annotations.append(f"__yq_tag_{i}_{v_node.tag}__")
            if isinstance(v_node, yaml.nodes.ScalarNode) and v_node.style:
                annotations.append(f"__yq_style_{i}_{v_node.style}__")
            elif (
                isinstance(v_node, yaml.nodes.SequenceNode | yaml.nodes.MappingNode)
                and v_node.flow_style is True
            ):
                annotations.append("__yq_style_{}_{}__".format(i, "flow"))
        return [loader.construct_object(i) for i in node.value] + annotations

    def construct_mapping(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> dict[Any, Any]:
        loader.flatten_mapping(node)  # TODO: is this needed?
        pairs = []
        for k_node, v_node in node.value:
            key = loader.construct_object(k_node)
            value = loader.construct_object(v_node)
            pairs.append((key, value))
            if not (use_annotations and isinstance(key, str | bytes)):
                continue
            if (
                v_node.tag
                and v_node.tag.startswith("!")
                and not v_node.tag.startswith("!!")
                and len(v_node.tag) > 1
            ):
                pairs.append((f"__yq_tag_{hash_key(key)}__", v_node.tag))
            if isinstance(v_node, yaml.nodes.ScalarNode) and v_node.style:
                pairs.append((f"__yq_style_{hash_key(key)}__", v_node.style))
            elif (
                isinstance(v_node, yaml.nodes.SequenceNode | yaml.nodes.MappingNode)
                and v_node.flow_style is True
            ):
                pairs.append((f"__yq_style_{hash_key(key)}__", "flow"))
        return dict(pairs)

    def parse_unknown_tags(loader: yaml.SafeLoader, tag_suffix: str, node: yaml.nodes.Node) -> Any:
        if isinstance(node, yaml.nodes.ScalarNode):
            return loader.construct_scalar(node)
        elif isinstance(node, yaml.nodes.SequenceNode):
            return construct_sequence(loader, node)
        elif isinstance(node, yaml.nodes.MappingNode):
            return construct_mapping(loader, node)

    loader_class = DefaultSafeLoader if expand_aliases else CustomLoader
    loader_class.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)
    loader_class.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, construct_sequence
    )
    loader_class.add_multi_constructor("", parse_unknown_tags)
    loader_class.yaml_constructors.pop("tag:yaml.org,2002:binary", None)
    loader_class.yaml_constructors.pop("tag:yaml.org,2002:set", None)
    set_yaml_grammar(loader_class, expand_merge_keys=expand_merge_keys)
    return loader_class



import dataclasses
import sys
from collections.abc import Callable
from typing import TYPE_CHECKING

from libresvip.core.tick_counter import skip_beat_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    ParamCurve,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.portamento import PortamentoPitch
from libresvip.utils.audio import audio_track_info
from libresvip.utils.music_math import (
    clamp,
    ratio_to_db,
)
from libresvip.utils.search import find_index, find_last_index

from .constants import TICK_RATE
from .interval_utils import ticks_to_position
from .model import (
    SVAudio,
    SVCurvePitchControl,
    SVMeter,
    SVMixer,
    SVNote,
    SVParamCurve,
    SVParameters,
    SVPhonemeAttribute,
    SVPoint,
    SVProject,
    SVTempo,
    SVTrack,
)
from .options import (
    OutputOptions,
    SVProjectVersionCompatibility,
    VibratoOption,
    synthv_language_presets,
)
from .phoneme_utils import default_phone_marks, number_of_phones, sv_g2p

if TYPE_CHECKING:
    from libresvip.model.point import Point


@dataclasses.dataclass
class SynthVGenerator:
    options: OutputOptions
    first_bar_tick: int = dataclasses.field(init=False)
    first_bar_tempo: list[SongTempo] = dataclasses.field(init=False)
    note_buffer: list[Note] = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    pitch_simulator: PitchSimulator = dataclasses.field(init=False)
    no_vibrato_indexes: set[int] = dataclasses.field(init=False)
    lyrics_phonemes: list[str] = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> SVProject:
        sv_project = SVProject()
        if self.options.version_compatibility == SVProjectVersionCompatibility.BELOW_1_9_0:
            sv_project.instant_mode_enabled = False
        elif (
            self.options.version_compatibility
            == SVProjectVersionCompatibility.BETWEEN_1_10_0_AND_1_11_2
        ):
            sv_project.version = SVProjectVersionCompatibility.BETWEEN_1_10_0_AND_1_11_2.value
        else:
            sv_project.version = SVProjectVersionCompatibility.ABOVE_2_0_0.value
        new_meters = skip_beat_list(project.time_signature_list, 1)
        self.first_bar_tick = round(project.time_signature_list[0].bar_length())
        self.first_bar_tempo = [
            tempo for tempo in project.song_tempo_list if tempo.position < self.first_bar_tick
        ]
        self.synchronizer = TimeSynchronizer(project.song_tempo_list, self.first_bar_tick)
        for tempo in self.synchronizer.tempo_list:
            sv_project.time_sig.tempo.append(self.generate_tempo(tempo))
        if any(
            (beat.denominator < 2 or beat.denominator > 16) for beat in project.time_signature_list
        ):
            sv_project.time_sig.meter.append(SVMeter(index=0, numerator=4, denominator=4))
        else:
            for meter in new_meters:
                sv_project.time_sig.meter.append(self.generate_meter(meter))
        for track_id, track in enumerate(project.track_list):
            if (sv_track := self.generate_track(track)) is not None:
                sv_track.disp_order = track_id
                sv_project.tracks.append(sv_track)
        return sv_project

    @staticmethod
    def generate_tempo(tempo: SongTempo) -> SVTempo:
        return SVTempo(position=ticks_to_position(tempo.position), bpm=tempo.bpm)

    @staticmethod
    def generate_meter(signature: TimeSignature) -> SVMeter:
        return SVMeter(
            index=signature.bar_index,
            numerator=signature.numerator,
            denominator=signature.denominator,
        )

    def generate_track(self, track: Track) -> SVTrack | None:
        sv_track = SVTrack(
            name=track.title,
            mixer=SVMixer(
                gain_decibel=self.generate_volume(track.volume),
                pan=track.pan,
                mute=track.mute,
                solo=track.solo,
            ),
        )
        if isinstance(track, SingingTrack):
            sv_track.main_ref.is_instrumental = False
            sv_track.disp_color = "ff7db235"
            language_preset = synthv_language_presets[self.options.language_override.value]
            sv_track.main_ref.database.language_override = language_preset.language
            sv_track.main_ref.database.phoneset_override = language_preset.phoneset
            if track.ai_singer_name:
                sv_track.main_ref.database.name = track.ai_singer_name
            self.note_buffer = track.note_list
            self.pitch_simulator = PitchSimulator(
                synchronizer=self.synchronizer,
                note_list=track.note_list,
                portamento=PortamentoPitch.sigmoid_portamento(),
            )
            sv_track.main_group.parameters = self.generate_params(track.edited_params)
            if self.options.version_compatibility == SVProjectVersionCompatibility.ABOVE_2_0_0:
                sv_track.main_group.pitch_controls = self.generate_pitch_controls(
                    track.edited_params.pitch.reduce_sample_rate(
                        self.options.down_sample, interrupt_value=-100
                    )
                )

            self.lyrics_phonemes = sv_g2p(
                (SVNote.normalize_phoneme(note) for note in track.note_list),
                [language_preset.language] * len(track.note_list),
            )

            sv_track.main_group.notes = self.generate_notes_with_phones(track.note_list)

            if self.options.vibrato == VibratoOption.NONE:
                for note in sv_track.main_group.notes:
                    note.attributes.vibrato_depth = 0
            elif self.options.vibrato == VibratoOption.HYBRID:
                for index in self.no_vibrato_indexes:
                    sv_track.main_group.notes[index].attributes.vibrato_depth = 0
        else:
            sv_track.main_ref.is_instrumental = True
            sv_track.disp_color = "ff4794cb"
            sv_track.main_ref.audio = SVAudio(filename=track.audio_file_path, duration=0)
            sv_track.main_ref.blick_offset = self.generate_audio_offset(track.offset)
            if (track_info := audio_track_info(track.audio_file_path)) is not None:
                sv_track.main_ref.audio.duration = track_info.duration / 1000
        return sv_track

    def generate_audio_offset(self, offset: int) -> int:
        if offset >= 0:
            return ticks_to_position(offset)
        current_pos = self.first_bar_tick
        actual_pos = self.first_bar_tick + offset
        res = 0.0
        i = len(self.first_bar_tempo) - 1
        for i in range(len(self.first_bar_tempo) - 1, -1, -1):
            if actual_pos > self.first_bar_tempo[i].position:
                break
            res -= (
                (current_pos - self.first_bar_tempo[i].position) * 120 / self.first_bar_tempo[i].bpm
            )
            current_pos = self.first_bar_tempo[i].position
        if i >= 0:
            res -= (current_pos - actual_pos) * 120 / self.first_bar_tempo[i].bpm
        else:
            res += actual_pos * 120 / self.first_bar_tempo[0].bpm
        return round(res * TICK_RATE)

    @staticmethod
    def generate_volume(volume: float) -> float:
        return max(ratio_to_db(max(volume, 0.06)), -24.0)

    def generate_params(self, parameters: Params) -> SVParameters:
        sv_params = SVParameters(
            loudness=self.generate_param_curve(
                parameters.volume.reduce_sample_rate(self.options.down_sample),
                0,
                0.0,
                lambda val: (
                    val / 1000.0 * 12.0
                    if val >= 0
                    else max(
                        ratio_to_db(val / 1000.0 + 1.0 if val > -997 else 0.0039),
                        -48.0,
                    )
                ),
            ),
            tension=self.generate_param_curve(
                parameters.strength.reduce_sample_rate(self.options.down_sample),
                0,
                0.0,
                (1000.0).__rtruediv__,
            ),
            breathiness=self.generate_param_curve(
                parameters.breath.reduce_sample_rate(self.options.down_sample),
                0,
                0.0,
                (1000.0).__rtruediv__,
            ),
        )
        if self.options.version_compatibility != SVProjectVersionCompatibility.ABOVE_2_0_0:
            sv_params.pitch_delta = self.generate_pitch_curve(
                parameters.pitch.reduce_sample_rate(self.options.down_sample, -100)
            )
        return sv_params

    def generate_pitch_controls(self, curve: ParamCurve) -> list[SVCurvePitchControl]:
        pitch_controls: list[SVCurvePitchControl] = []
        buffer: list[Point] = []
        for point in curve.points.root:
            if point.x >= self.first_bar_tick:
                point = point._replace(x=point.x - self.first_bar_tick)
                if point.y == -100:
                    if not buffer:
                        continue
                    base_positon = ticks_to_position(buffer[0].x)
                    base_pitch = buffer[0].y
                    pitch_controls.append(
                        SVCurvePitchControl(
                            pos=base_positon,
                            pitch=base_pitch / 100.0,
                            id_value=str(len(pitch_controls) + 1),
                            points=[
                                SVPoint(
                                    offset=ticks_to_position(tmp_point.x) - base_positon,
                                    value=(tmp_point.y - base_pitch) / 100.0,
                                )
                                for tmp_point in buffer
                            ],
                        )
                    )
                    buffer.clear()
                else:
                    buffer.append(point)
        if buffer:
            base_positon = ticks_to_position(buffer[0].x)
            base_pitch = buffer[0].y
            pitch_controls.append(
                SVCurvePitchControl(
                    pos=base_positon,
                    pitch=base_pitch / 100.0,
                    id_value=str(len(pitch_controls) + 1),
                    points=[
                        SVPoint(
                            offset=ticks_to_position(tmp_point.x) - base_positon,
                            value=(tmp_point.y - base_pitch) / 100.0,
                        )
                        for tmp_point in buffer
                    ],
                )
            )
        return pitch_controls

    def generate_pitch_curve(self, curve: ParamCurve) -> SVParamCurve:
        sv_curve = SVParamCurve()
        if self.options.vibrato == VibratoOption.HYBRID:
            self.no_vibrato_indexes = set()
        if not len(self.note_buffer):
            return sv_curve
        point_list = sv_curve.points
        buffer: list[Point] = []
        min_interval = 1
        last_point = None
        for point in curve.points.root:
            if point.x >= self.first_bar_tick:
                point = point._replace(x=point.x - self.first_bar_tick)
                if point.y == -100:
                    if not buffer:
                        continue
                    if last_point is None or last_point.x + min_interval < buffer[0].x:
                        if last_point is not None and last_point.x + 2 * min_interval < buffer[0].x:
                            point_list.append(
                                SVPoint(
                                    offset=ticks_to_position(last_point.x + min_interval),
                                    value=0,
                                )
                            )
                        point_list.append(
                            SVPoint(
                                offset=ticks_to_position(buffer[0].x - min_interval),
                                value=0,
                            )
                        )
                    for tmp_point in buffer:
                        point_list.append(
                            SVPoint(
                                offset=ticks_to_position(tmp_point.x),
                                value=self.generate_pitch_diff(tmp_point.x, tmp_point.y),
                            )
                        )
                    last_point = buffer[-1]
                    buffer.clear()
                else:
                    buffer.append(point)
        if last_point is not None:
            point_list.append(
                SVPoint(
                    offset=ticks_to_position(last_point.x + min_interval),
                    value=0,
                )
            )
        return sv_curve

    def generate_note(self, note: Note) -> SVNote:
        onset = ticks_to_position(note.start_pos)
        return SVNote(
            instant_mode=False
            if self.options.version_compatibility
            == SVProjectVersionCompatibility.BETWEEN_1_10_0_AND_1_11_2
            else None,
            onset=onset,
            pitch=note.key_number,
            lyrics=note.lyric or note.pronunciation or "",
            duration=ticks_to_position(note.end_pos) - onset,
        )

    def generate_pitch_diff(self, pos: int, pitch: int) -> float:
        target_note_index = find_last_index(self.note_buffer, lambda x: x.start_pos <= pos)
        target_note = self.note_buffer[target_note_index] if target_note_index >= 0 else None
        if (
            simulated_pitch := self.pitch_simulator.pitch_at_secs(
                self.synchronizer.get_actual_secs_from_ticks(pos)
            )
        ) is not None:
            pitch_diff = pitch - simulated_pitch
        else:
            pitch_diff = 0.0
        if target_note is None:
            return pitch_diff
        if (
            self.options.vibrato == VibratoOption.HYBRID
            and self.synchronizer.get_duration_secs_from_ticks(target_note.start_pos, pos) > 0.25
            and pos < target_note.end_pos
        ):
            self.no_vibrato_indexes.add(target_note_index)
        return pitch_diff

    def generate_param_curve(
        self,
        curve: ParamCurve,
        termination: int,
        default_value: float,
        mapping_func: Callable[[int], float],
    ) -> SVParamCurve:
        sv_curve = SVParamCurve()
        if not len(curve.points):
            return sv_curve
        if self.options.down_sample > 15:
            sv_curve.mode = "cubic"
        skipped = 0
        point_list = sv_curve.points
        if curve.points[0].x == -192000:
            if len(curve.points) == 2 and curve.points[1].x == sys.maxsize // 2:
                if curve.points[0].y != termination:
                    point_list.append(SVPoint(offset=0, value=mapping_func(curve.points[0].y)))
                return sv_curve
            skipped = 1
            valid_index = find_index(curve.points.root, lambda x: x.x >= self.first_bar_tick)
            if (
                valid_index != -1
                and len(curve.points) > valid_index + 1
                and (
                    curve.points[valid_index].y != termination
                    or curve.points[valid_index + 1].y != termination
                    or curve.points[valid_index + 1].x >= sys.maxsize // 2
                )
            ):
                skipped = valid_index + 1
                x0 = curve.points[valid_index].x
                y0 = curve.points[valid_index].y
                point_list.append(
                    SVPoint(
                        offset=ticks_to_position(x0 - self.first_bar_tick),
                        value=mapping_func(y0),
                    )
                )
        buffer: list[Point] = []
        min_interval = 1
        last_point = None
        for point in curve.points.root[skipped:]:
            if self.first_bar_tick <= point.x < sys.maxsize // 2:
                point = point._replace(x=point.x - self.first_bar_tick)
                if point.y == termination:
                    if not buffer:
                        continue
                    if last_point is None or last_point.x + min_interval < buffer[0].x:
                        if last_point is not None and last_point.x + 2 * min_interval < buffer[0].x:
                            point_list.append(
                                SVPoint(
                                    offset=ticks_to_position(last_point.x + min_interval),
                                    value=default_value,
                                )
                            )
                        point_list.append(
                            SVPoint(
                                offset=ticks_to_position(buffer[0].x - min_interval),
                                value=default_value,
                            )
                        )
                    for tmp_point in buffer:
                        point_list.append(
                            SVPoint(
                                offset=ticks_to_position(tmp_point.x),
                                value=mapping_func(tmp_point.y),
                            )
                        )
                    last_point = buffer[-1]
                    buffer.clear()
                else:
                    buffer.append(point)

        if not buffer:
            if last_point is not None:
                point_list.append(
                    SVPoint(
                        offset=ticks_to_position(last_point.x + min_interval),
                        value=default_value,
                    )
                )
            return sv_curve
        if last_point is None or last_point.x + min_interval < buffer[0].x:
            if last_point is not None and last_point.x + 2 * min_interval < buffer[0].x:
                point_list.append(
                    SVPoint(
                        offset=ticks_to_position(last_point.x + min_interval),
                        value=default_value,
                    )
                )
            point_list.append(
                SVPoint(
                    offset=ticks_to_position(buffer[0].x - min_interval),
                    value=default_value,
                )
            )
        for tmp_point in buffer:
            point_list.append(
                SVPoint(
                    offset=ticks_to_position(tmp_point.x),
                    value=mapping_func(tmp_point.y),
                )
            )
        last_point = buffer[-1]
        buffer.clear()
        if last_point.y == termination:
            point_list.append(
                SVPoint(
                    offset=ticks_to_position(last_point.x + min_interval),
                    value=default_value,
                )
            )
        return sv_curve

    def generate_notes_with_phones(self, notes: list[Note]) -> list[SVNote]:
        sv_note_list: list[SVNote] = []
        if not len(notes):
            return sv_note_list
        current_note = notes[0]
        current_sv_note = self.generate_note(current_note)
        current_phone_marks = default_phone_marks(
            self.lyrics_phonemes[0], self.options.language_override.value
        )
        if (
            current_note.edited_phones is not None
            and current_note.edited_phones.head_length_in_secs > 0
        ):
            if self.options.version_compatibility == SVProjectVersionCompatibility.ABOVE_2_0_0:
                current_sv_note.attributes.phonemes = [
                    SVPhonemeAttribute(left_offset=-current_note.edited_phones.head_length_in_secs)
                ]
            elif current_phone_marks[0] > 0:
                ratio = current_note.edited_phones.head_length_in_secs / current_phone_marks[0]
                current_sv_note.attributes.set_phone_duration(0, clamp(ratio, 0.2, 1.8))
        for next_note, cur_phoneme, next_phoneme in zip(
            notes[1:], self.lyrics_phonemes[:-1], self.lyrics_phonemes[1:]
        ):
            next_sv_note = self.generate_note(next_note)
            next_phone_marks = default_phone_marks(
                next_phoneme, self.options.language_override.value
            )

            current_main_part_edited = (
                current_phone_marks[1] > 0
                and current_note.edited_phones is not None
                and current_note.edited_phones.mid_ratio_over_tail > 0
            )
            next_head_part_edited = (
                next_phone_marks[0] > 0
                and next_note.edited_phones is not None
                and next_note.edited_phones.head_length_in_secs > 0
            )

            index = 1 if current_phone_marks[0] > 0 else 0
            if (
                next_note.edited_phones is not None
                and next_note.edited_phones.head_length_in_secs > 0
            ):
                next_sv_note.attributes.phonemes = [
                    SVPhonemeAttribute(left_offset=-next_note.edited_phones.head_length_in_secs)
                ]
            elif (
                current_main_part_edited
                and next_head_part_edited
                and (current_note.edited_phones is not None)
                and (next_note.edited_phones is not None)
            ):
                current_main_ratio = (
                    current_note.edited_phones.mid_ratio_over_tail / current_phone_marks[1]
                )
                next_head_ratio = next_note.edited_phones.head_length_in_secs / next_phone_marks[0]
                x = 2 * current_main_ratio / (1 + current_main_ratio)
                y = 2 / (1 + current_main_ratio)
                z = next_head_ratio
                if (
                    self.synchronizer.get_duration_secs_from_ticks(
                        current_note.end_pos, next_note.start_pos
                    )
                    < next_phone_marks[0]
                ):
                    final_ratio = 2 / (1 + next_head_ratio)
                    x *= final_ratio
                    y *= final_ratio
                    z *= final_ratio
                current_sv_note.attributes.set_phone_duration(index, clamp(x, 0.2, 1.8))
                current_sv_note.attributes.set_phone_duration(index + 1, clamp(y, 0.2, 1.8))
                next_sv_note.attributes.set_phone_duration(0, clamp(z, 0.2, 1.8))
            elif current_main_part_edited and current_note.edited_phones is not None:
                ratio = current_note.edited_phones.mid_ratio_over_tail / current_phone_marks[1]
                x = 2 * ratio / (1 + ratio)
                y = 2 / (1 + ratio)
                current_sv_note.attributes.set_phone_duration(index, clamp(x, 0.2, 1.8))
                current_sv_note.attributes.set_phone_duration(index + 1, clamp(y, 0.2, 1.8))
            elif next_head_part_edited and next_note.edited_phones is not None:
                ratio = next_note.edited_phones.head_length_in_secs / next_phone_marks[0]
                if (
                    self.synchronizer.get_duration_secs_from_ticks(
                        current_note.end_pos, next_note.start_pos
                    )
                    < next_phone_marks[0]
                ):
                    ratio_z = 2 * ratio / (1 + ratio)
                    ratio_xy = 2 / (1 + ratio)
                    ratio_z = clamp(ratio_z, 0.2, 1.8)
                    ratio_xy = clamp(ratio_xy, 0.2, 1.8)
                    current_sv_note.attributes.set_phone_duration(index, ratio_xy)
                    if current_phone_marks[1] > 0:
                        current_sv_note.attributes.set_phone_duration(index + 1, ratio_xy)
                    ratio = ratio_z
                next_sv_note.attributes.set_phone_duration(0, ratio)
            if current_sv_note.attributes.dur is not None:
                expected_length = number_of_phones(
                    cur_phoneme, self.options.language_override.value
                )
                if len(current_sv_note.attributes.dur) < expected_length:
                    current_sv_note.attributes.set_phone_duration(expected_length - 1, 1.0)

            sv_note_list.append(current_sv_note)

            current_note = next_note
            current_sv_note = next_sv_note
            current_phone_marks = next_phone_marks
        if (
            current_phone_marks[1] > 0
            and current_note.edited_phones is not None
            and current_note.edited_phones.mid_ratio_over_tail > 0
            and notes[-1].edited_phones is not None
        ):
            ratio = notes[-1].edited_phones.mid_ratio_over_tail / current_phone_marks[1]
            x = 2 * ratio / (1 + ratio)
            y = 2 / (1 + ratio)
            index = 1 if current_phone_marks[0] > 0 else 0
            current_sv_note.attributes.set_phone_duration(index, clamp(x, 0.2, 1.8))
            current_sv_note.attributes.set_phone_duration(index + 1, clamp(y, 0.2, 1.8))
        if current_sv_note.attributes.dur is not None:
            expected_length = number_of_phones(
                self.lyrics_phonemes[-1], self.options.language_override.value
            )
            if len(current_sv_note.attributes.dur) < expected_length:
                current_sv_note.attributes.set_phone_duration(expected_length - 1, 1.0)
        sv_note_list.append(current_sv_note)
        return sv_note_list



import dataclasses

from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import legato_chars
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import (
    PpsfAudioTrackItem,
    PpsfCurveType,
    PpsfDvlTrackEvent,
    PpsfDvlTrackItem,
    PpsfEventTrack,
    PpsfMeters,
    PpsfProject,
    PpsfTempos,
)
from .options import InputOptions
from .ppsf_interval_dict import ppsf_key_interval_dict


@dataclasses.dataclass
class PiaproStudioNTParser:
    options: InputOptions
    first_bar_ticks: int = dataclasses.field(init=False)

    def parse_project(self, ppsf_project: PpsfProject) -> Project:
        time_signatures = self.parse_time_signatures(ppsf_project.ppsf.project.meter)
        self.first_bar_ticks = round(time_signatures[0].bar_length())
        tempos = self.parse_tempos(ppsf_project.ppsf.project.tempo)
        singing_tracks = self.parse_singing_tracks(
            ppsf_project.ppsf.project.dvl_track,
            ppsf_project.ppsf.gui_settings.track_editor.event_tracks,
        )
        instrumental_tracks = self.parse_instrumental_tracks(ppsf_project.ppsf.project.audio_track)
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=singing_tracks + instrumental_tracks,
        )

    def parse_time_signatures(self, ppsf_meters: PpsfMeters) -> list[TimeSignature]:
        time_signatures: list[TimeSignature] = []
        first_time_signature = TimeSignature(
            numerator=ppsf_meters.const.nume,
            denominator=ppsf_meters.const.denomi,
        )
        if ppsf_meters.use_sequence:
            time_signatures.extend(
                TimeSignature(
                    numerator=meter.nume,
                    denominator=meter.denomi,
                    bar_index=meter.measure,
                )
                for meter in ppsf_meters.sequence
            )
        if not time_signatures or time_signatures[0].bar_index != 0:
            time_signatures.insert(0, first_time_signature)
        return time_signatures

    def parse_tempos(self, ppsf_tempos: PpsfTempos) -> list[SongTempo]:
        tempos: list[SongTempo] = []
        first_tempo = SongTempo(bpm=ppsf_tempos.const / 10000, position=0)
        if ppsf_tempos.use_sequence:
            tempos.extend(
                SongTempo(bpm=tempo.value / 10000, position=tempo.tick)
                for tempo in ppsf_tempos.sequence
            )
        if not tempos or tempos[0].position != 0:
            tempos.insert(0, first_tempo)
        return tempos

    def parse_instrumental_tracks(
        self, ppsf_audio_tracks: list[PpsfAudioTrackItem]
    ) -> list[InstrumentalTrack]:
        tracks = []
        if self.options.import_instrumental_track:
            for track in ppsf_audio_tracks:
                for i, event in enumerate(track.events):
                    instrumental_track = InstrumentalTrack(
                        title=f"{track.name} {i + 1}",
                        audio_file_path=event.file_audio_data.file_path,
                        offset=event.tick_pos,
                    )
                    tracks.append(instrumental_track)
        return tracks

    def parse_singing_tracks(
        self,
        ppsf_dvl_tracks: list[PpsfDvlTrackItem] | None,
        event_tracks: list[PpsfEventTrack],
    ) -> list[SingingTrack]:
        tracks = []
        if ppsf_dvl_tracks is not None:
            for track, event_track in zip(ppsf_dvl_tracks, event_tracks):
                singing_track = SingingTrack(
                    title=track.name,
                    ai_singer_name=track.singer.singer_name,
                    note_list=self.parse_notes(track.events),
                    solo=event_track.mute_solo > 0,
                    mute=event_track.mute_solo < 0,
                )
                for parameter in track.parameters:
                    if (
                        self.options.import_pitch
                        and parameter.base_sequence is not None
                        and parameter.base_sequence.name == "pitch_bend"
                    ):
                        key_interval_dict = ppsf_key_interval_dict(track.events, event_track.notes)
                        for point in parameter.base_sequence.sequence:
                            if point.curve_type == PpsfCurveType.BORDER and point.value == 0:
                                singing_track.edited_params.pitch.points.append(
                                    Point(
                                        x=point.pos + self.first_bar_ticks,
                                        y=-100,
                                    )
                                )
                            elif (base_key := key_interval_dict.get(point.pos)) is not None:
                                singing_track.edited_params.pitch.points.append(
                                    Point(
                                        x=point.pos + self.first_bar_ticks,
                                        y=point.value + round(base_key * 100),
                                    )
                                )
                tracks.append(singing_track)
        return tracks

    def parse_notes(self, ppsf_dvl_track_events: list[PpsfDvlTrackEvent]) -> list[Note]:
        return [
            Note(
                key_number=event.note_number,
                start_pos=event.pos,
                length=event.length,
                lyric="-" if event.lyric in legato_chars else event.lyric,
                pronunciation=event.symbols,
            )
            for event in ppsf_dvl_track_events
            if event.enabled
        ]



# mypy: disable-error-code="misc"
import enum
from dataclasses import dataclass
from typing import Annotated, Literal

from pydantic import ConfigDict, Field, ValidationInfo, model_validator
from typing_extensions import Self

from libresvip.model.base import BaseModel
from libresvip.utils.translation import gettext_lazy as _


class VocaloidLanguage(enum.IntEnum):
    JAPANESE: Annotated[int, Field(title=_("日本語"))] = 0
    ENGLISH: Annotated[int, Field(title=_("English"))] = 1
    KOREAN: Annotated[int, Field(title=_("한국어"))] = 2
    SPANISH: Annotated[int, Field(title=_("Español"))] = 3
    SIMPLIFIED_CHINESE: Annotated[int, Field(title=_("简体中文"))] = 4


class VocaloidTrackType(enum.IntEnum):
    STANDARD = 0
    AUDIO = 1
    AI = 2


@dataclass
class ControllerEvent:
    pos: int
    value: int


@dataclass
class VocaloidPartPitchData:
    start_pos: int
    pit: list[ControllerEvent]
    pbs: list[ControllerEvent]


class VocaloidBasePos(BaseModel):
    pos: int = 0


class VocaloidWithDur(VocaloidBasePos):
    duration: int | None = None


class VocaloidPoint(VocaloidBasePos):
    value: int | float | Literal["ZeroPitch"] | None = None


class VocaloidTimeSig(BaseModel):
    bar: int = 0
    denom: int = 4
    numer: int = 4


class VocaloidExp(BaseModel):
    opening: int | None = 127
    accent: int | None = None
    decay: int | None = None
    bend_depth: int | None = Field(None, alias="bendDepth")
    bend_length: int | None = Field(None, alias="bendLength")


class VocaloidAIExp(BaseModel):
    pitch_fine: float | None = Field(None, alias="pitchFine")
    pitch_drift_start: float | None = Field(None, alias="pitchDriftStart")
    pitch_drift_end: float | None = Field(None, alias="pitchDriftEnd")
    pitch_scaling_center: float | None = Field(None, alias="pitchScalingCenter")
    pitch_scaling_origin: float | None = Field(None, alias="pitchScalingOrigin")
    pitch_transition_start: float | None = Field(None, alias="pitchTransitionStart")
    pitch_transition_end: float | None = Field(None, alias="pitchTransitionEnd")
    amplitude_whole: float | None = Field(None, alias="amplitudeWhole")
    amplitude_start: float | None = Field(None, alias="amplitudeStart")
    amplitude_end: float | None = Field(None, alias="amplitudeEnd")
    formant_whole: float | None = Field(None, alias="formantWhole")
    formant_start: float | None = Field(None, alias="formantStart")
    formant_end: float | None = Field(None, alias="formantEnd")
    vibrato_leading_depth: float | None = Field(None, alias="vibratoLeadingDepth")
    vibrato_following_depth: float | None = Field(None, alias="vibratoFollowingDepth")


class VocaloidEnabled(BaseModel):
    is_enabled: bool | None = Field(True, alias="isEnabled")


class VocaloidGlobal(VocaloidEnabled):
    value: int | None = 12000


class VocaloidRegion(VocaloidEnabled):
    begin: float | None = 0
    end: float | None = 7680


class VocaloidParameters(BaseModel):
    name: str | None = None
    value: int | float | str | None = None


class VocaloidVersion(BaseModel):
    major: int = 5
    minor: int = 0
    revision: int = 0


class VocaloidVibrato(BaseModel):
    type_value: int | None = Field(0, alias="type")
    duration: int | None = 0
    depths: list[VocaloidPoint] | None = Field(default_factory=list)
    rates: list[VocaloidPoint] | None = Field(default_factory=list)


class VocaloidCompID(BaseModel):
    comp_id: str | None = Field(None, alias="compID")


class VocaloidLangID(BaseModel):
    lang_id: VocaloidLanguage | None = Field(VocaloidLanguage.SIMPLIFIED_CHINESE, alias="langID")


class VocaloidVoice(VocaloidCompID, VocaloidLangID):
    pass


class VocaloidAIVoice(VocaloidCompID):
    lang_ids: list[VocaloidLangID] = Field(default_factory=list, alias="langIDs")


class VocaloidVoices(VocaloidCompID):
    name: str | None = None


class VocaloidWeight(BaseModel):
    pre: int | None = 64
    post: int | None = 64


class VocaloidControllers(BaseModel):
    name: str | None = None
    events: list[VocaloidPoint] = Field(default_factory=list)


class VocaloidFolded(BaseModel):
    is_folded: bool | None = Field(True, alias="isFolded")


class VocaloidEffects(VocaloidFolded):
    id_value: str | None = Field(None, alias="id")
    is_bypassed: bool | None = Field(None, alias="isBypassed")
    parameters: list[VocaloidParameters] = Field(default_factory=list)


class VocaloidSingingSkill(BaseModel):
    duration: int | None = 158
    weight: VocaloidWeight | None = Field(default_factory=VocaloidWeight)


class VocaloidTempo(VocaloidFolded):
    events: list[VocaloidPoint] = Field(default_factory=list)
    global_value: VocaloidGlobal | None = Field(alias="global", default_factory=VocaloidGlobal)
    height: int | None = 0
    ara: VocaloidEnabled | None = None


class VocaloidTimeSigs(VocaloidFolded):
    events: list[VocaloidTimeSig] = Field(default_factory=list)


class VocaloidAutomation(VocaloidFolded):
    height: int = 0
    events: list[VocaloidPoint] = Field(default_factory=list)


class VocaloidMasterTrack(BaseModel):
    loop: VocaloidRegion = Field(default_factory=VocaloidRegion)
    sampling_rate: int = Field(44100, alias="samplingRate")
    tempo: VocaloidTempo = Field(default_factory=VocaloidTempo)
    time_sig: VocaloidTimeSigs = Field(alias="timeSig", default_factory=VocaloidTimeSigs)
    volume: VocaloidAutomation = Field(default_factory=VocaloidAutomation)
    main_tuning: float = Field(440, alias="mainTuning")


class VocaloidDVQMRelease(VocaloidCompID):
    speed: int | None = None
    level_names: list[str] | None = Field(None, alias="levelNames")
    top_factor: float | None = Field(None, alias="topFactor")
    is_protected: bool | None = Field(None, alias="isProtected")


class VocaloidDVQM(BaseModel):
    release: VocaloidDVQMRelease | None = None
    attack: VocaloidDVQMRelease | None = None


class VocaloidNotes(VocaloidLangID, VocaloidWithDur):
    exp: VocaloidExp | None = Field(default_factory=VocaloidExp)
    ai_exp: VocaloidAIExp | None = Field(None, alias="aiExp")
    direct_pitches: list[VocaloidPoint] | None = Field(None, alias="directPitches")
    is_protected: bool | None = Field(False, alias="isProtected")
    is_ai_vibrato_enabled: bool | None = Field(False, alias="isAiVibratoEnabled")
    lyric: str
    number: int
    phoneme: str | None = None
    phoneme_positions: list[VocaloidBasePos] | None = Field(None, alias="phonemePositions")
    singing_skill: VocaloidSingingSkill | None = Field(
        alias="singingSkill", default_factory=VocaloidSingingSkill
    )
    velocity: int = 64
    vibrato: VocaloidVibrato | None = Field(default_factory=VocaloidVibrato)
    dvqm: VocaloidDVQM | None = None


class VocaloidWav(BaseModel):
    name: str
    original_name: str | None = Field(None, alias="originalName")

    @model_validator(mode="after")
    def extract_audio(self, info: ValidationInfo) -> Self:
        if (
            info.context is not None
            and info.context["extract_audio"]
            and not hasattr(info.context["path"], "protocol")
        ):
            archive_wav_path = f"Project/Audio/{self.name}"
            if not (
                wav_path := (info.context["path"].parent / (self.original_name or self.name))
            ).exists():
                wav_path.write_bytes(info.context["archive_file"].read(archive_wav_path))
        return self


class VocaloidVoicePart(VocaloidWithDur):
    name: str | None = ""
    midi_effects: list[VocaloidEffects] = Field(default_factory=list, alias="midiEffects")
    audio_effects: list[VocaloidEffects] = Field(default_factory=list, alias="audioEffects")
    notes: list[VocaloidNotes] = Field(default_factory=list)
    style_preset_id: str | None = Field(None, alias="stylePresetID")
    style_name: str | None = Field("No Effect", alias="styleName")
    voice: VocaloidVoice | None = None
    ai_voice: VocaloidAIVoice | None = Field(None, alias="aiVoice")
    controllers: list[VocaloidControllers] | None = None

    def get_controller_events(self, name: str) -> list[ControllerEvent]:
        if self.controllers is None:
            return []
        return [
            ControllerEvent(
                pos=event.pos,
                value=int(event.value),
            )
            for controller in self.controllers
            if controller.name == name
            for event in controller.events
            if event.value is not None
        ]


class VocaloidWavPart(VocaloidBasePos):
    model_config = ConfigDict(
        populate_by_name=True,
        extra="forbid",
    )
    region: VocaloidRegion | None = None
    name: str | None = ""
    wav: VocaloidWav | None = None


class VocaloidBaseTracks(VocaloidFolded):
    bus_no: int | None = Field(0, alias="busNo")
    color: int | None = 0
    height: float | None = 0
    is_muted: bool = Field(False, alias="isMuted")
    is_solo_mode: bool = Field(False, alias="isSoloMode")
    name: str | None = ""
    panpot: VocaloidAutomation = Field(default_factory=VocaloidAutomation)
    volume: VocaloidAutomation = Field(default_factory=VocaloidAutomation)
    last_scroll_position_note_number: int | None = Field(None, alias="lastScrollPositionNoteNumber")


class VocaloidStandardTrack(VocaloidBaseTracks):
    parts: list[VocaloidVoicePart] = Field(default_factory=list)
    type_value: Literal[VocaloidTrackType.STANDARD] = Field(
        VocaloidTrackType.STANDARD, alias="type"
    )


class VocaloidAITrack(VocaloidBaseTracks):
    parts: list[VocaloidVoicePart] = Field(default_factory=list)
    type_value: Literal[VocaloidTrackType.AI] = Field(VocaloidTrackType.AI, alias="type")


class VocaloidAudioTrack(VocaloidBaseTracks):
    parts: list[VocaloidWavPart] = Field(default_factory=list)
    type_value: Literal[VocaloidTrackType.AUDIO] = Field(VocaloidTrackType.AUDIO, alias="type")


VocaloidTracks = Annotated[
    VocaloidStandardTrack | VocaloidAITrack | VocaloidAudioTrack,
    Field(discriminator="type_value"),
]


class VocaloidProject(BaseModel):
    master_track: VocaloidMasterTrack = Field(
        alias="masterTrack", default_factory=VocaloidMasterTrack
    )
    title: str | None = ""
    tracks: list[VocaloidTracks] = Field(default_factory=list)
    vender: str = "Yamaha Corporation"
    version: VocaloidVersion | None = Field(default_factory=VocaloidVersion)
    voices: list[VocaloidVoices] = Field(default_factory=list)



from types import SimpleNamespace
from typing import Final

TICK_RATE: Final[int] = 1470000
DEFAULT_PITCH_TRANSITION: Final[float] = 0.0
DEFAULT_PITCH_PORTAMENTO: Final[float] = 0.07
DEFAULT_PITCH_DEPTH: Final[float] = 0.15
DEFAULT_VIBRATO_START: Final[float] = 0.25
DEFAULT_VIBRATO_FADE: Final[float] = 0.2
DEFAULT_VIBRATO_DEPTH: Final[float] = 1.0
DEFAULT_VIBRATO_FREQUENCY: Final[float] = 5.5
DEFAULT_VIBRATO_PHASE: Final[float] = 0.0
DEFAULT_VIBRATO_JITTER: Final[float] = 1.0
SYSTEM_PITCH_PORTAMENTO: Final[float] = 0.1
SYSTEM_PITCH_DEPTH: Final[float] = 0.0
MAX_BREAK: Final[float] = 0.01
DEFAULT_PHONE_RATIO: Final[float] = 1.8
DEFAULT_DURATIONS = SimpleNamespace(
    stop=0.10,
    affricate=0.125,
    fricative=0.125,
    aspirate=0.094,
    liquid=0.062,
    nasal=0.094,
    vowel=0.0,
    semivowel=0.055,
    diphthong=0.0,
)



import dataclasses
from typing import cast

from construct_typed import DataclassMixin, DataclassStruct

from libresvip.core.lyric_phoneme.chinese import CHINESE_RE, get_pinyin_series
from libresvip.core.tick_counter import shift_beat_list, shift_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.audio import audio_track_info

from .constants import (
    DEFAULT_VOLUME,
    MIN_SEGMENT_LENGTH,
)
from .deepvocal_pitch import convert_note_key, generate_for_dv
from .model import (
    DvAudioInfo,
    DvAudioTrack,
    DvInnerProject,
    DvNote,
    DvNoteParameter,
    DvPhoneme,
    DvPoint,
    DvProject,
    DvSegment,
    DvSingingTrack,
    DvTempo,
    DvTimeSignature,
    DvTrack,
    DvTrackType,
)
from .options import OutputOptions


def size_of_object(obj: DataclassMixin) -> int:
    return len(DataclassStruct(type(obj)).build(obj))


@dataclasses.dataclass
class DeepVocalGenerator:
    options: OutputOptions
    tick_prefix: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> DvProject:
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        self.tick_prefix = round(4 * project.time_signature_list[0].bar_length())
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        singing_tracks = self.generate_singing_tracks(
            [track for track in project.track_list if isinstance(track, SingingTrack)]
        )
        audio_tracks = self.generate_instrumental_tracks(
            [track for track in project.track_list if isinstance(track, InstrumentalTrack)]
        )
        return DvProject(
            inner_project=DvInnerProject(
                tracks=singing_tracks + audio_tracks,
                tempos=self.generate_tempos(project.song_tempo_list),
                time_signatures=self.generate_time_signatures(project.time_signature_list),
            )
        )

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature]
    ) -> list[DvTimeSignature]:
        return [
            DvTimeSignature(
                measure_position=ts.bar_index,
                numerator=ts.numerator,
                denominator=ts.denominator,
            )
            for ts in shift_beat_list(
                [
                    time_sig if i > 0 else time_sig.model_copy(update={"bar_index": -3})
                    for i, time_sig in enumerate(time_signatures)
                ],
                1,
            )
        ]

    def generate_tempos(self, tempos: list[SongTempo]) -> list[DvTempo]:
        return [
            DvTempo(
                position=tempo.position,
                bpm=round(tempo.bpm * 100),
            )
            for tempo in shift_tempo_list(
                [
                    song_tempo if i > 0 else song_tempo.model_copy(update={"position": 0})
                    for i, song_tempo in enumerate(tempos)
                ],
                self.tick_prefix - self.first_bar_length,
            )
        ]

    def generate_instrumental_tracks(
        self, instrumental_tracks: list[InstrumentalTrack]
    ) -> list[DvTrack]:
        track_list = []
        for track in instrumental_tracks:
            if (track_info := audio_track_info(track.audio_file_path)) is not None:
                audio_info = DvAudioInfo(
                    path=track.audio_file_path,
                    name=track.title,
                    start=track.offset + self.tick_prefix,
                    length=round(
                        self.time_synchronizer.get_actual_ticks_from_secs_offset(
                            track.offset, track_info.duration / 1000
                        )
                    ),
                )
                dv_track = DvAudioTrack(
                    name=track.title,
                    mute=track.mute,
                    solo=track.solo,
                    volume=DEFAULT_VOLUME,
                    balance=0,
                    infos=[audio_info],
                )
                track_list.append(
                    DvTrack(
                        track_type=cast("DvTrackType", DvTrackType.AUDIO),
                        track_data=dv_track,
                    )
                )
        return track_list

    def generate_singing_tracks(self, singing_tracks: list[SingingTrack]) -> list[DvTrack]:
        track_list = []
        for track in singing_tracks:
            dv_notes = self.generate_notes(track.note_list)
            dv_segment = DvSegment(
                start=self.tick_prefix,
                name=track.title,
                singer_name=track.ai_singer_name,
                length=max(
                    (note.end_pos for note in track.note_list),
                    default=MIN_SEGMENT_LENGTH,
                ),
                notes=dv_notes,
                volume_data=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
                pitch_data=[
                    DvPoint(x=-1, y=-1),
                    DvPoint(x=307201, y=-1),
                ],
                unknown_1=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
                breath_data=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
                gender_data=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
                unknown_2=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
                unknown_3=[
                    DvPoint(x=-1, y=128),
                    DvPoint(x=307201, y=128),
                ],
            )
            if (
                dv_segment_pitch_raw_data := generate_for_dv(
                    self.first_bar_length,
                    track.edited_params.pitch,
                    track.note_list,
                )
            ) is not None:
                dv_segment.pitch_data = dv_segment_pitch_raw_data.data
            dv_track = DvSingingTrack(
                name=track.title,
                mute=track.mute,
                solo=track.solo,
                volume=DEFAULT_VOLUME,
                balance=0,
                segments=[dv_segment],
            )
            track_list.append(
                DvTrack(
                    track_type=cast("DvTrackType", DvTrackType.SINGING),
                    track_data=dv_track,
                )
            )
        return track_list

    def generate_notes(self, notes: list[Note]) -> list[DvNote]:
        return [
            DvNote(
                start=note.start_pos,
                length=note.length,
                key=cast("int", convert_note_key(note.key_number)),
                phoneme=cast("str", note.pronunciation)
                or (
                    note.lyric
                    if CHINESE_RE.search(note.lyric) is None
                    else next(iter(get_pinyin_series(note.lyric)), "")
                ),
                word=note.lyric,
                padding_1=0,
                vibrato=50,
                note_vibrato_data=DvNoteParameter(
                    amplitude_points=[
                        DvPoint(x=-1, y=0),
                        DvPoint(x=100001, y=0),
                    ],
                    frequency_points=[
                        DvPoint(x=-1, y=0),
                        DvPoint(x=100001, y=0),
                    ],
                    vibrato_points=[
                        DvPoint(x=0, y=0),
                        DvPoint(x=1124, y=0),
                    ],
                ),
                phonemes=DvPhoneme(
                    unknown_1=0,
                    consonant_rate=1.0,
                    vowel_modified=0,
                    medial=1.0,
                    rime=1.0,
                    ending=1.0,
                ),
                ben_depth=0,
                ben_length=0,
                por_head=0,
                por_tail=0,
                timbre=-1,
                cross_lyric="",
                cross_timbre=-1,
            )
            for note in notes
        ]



from types import SimpleNamespace
from typing import Final

TYPE_URL_BASE: Final[str] = "type.googleapis.com/"
DEFAULT_SINGER_ID: Final[str] = "7d0c0cfc-00b3-4dca-b7b0-d20b634b531a"
MIN_NOTE_DURATION: Final[float] = 0.06
MAX_NOTE_DURATION: Final[float] = 6.0


Svip3TrackType = SimpleNamespace(
    SINGING_TRACK="xstudio.proto.SingingTrack",
    AUDIO_TRACK="xstudio.proto.AudioTrack",
)



from __future__ import annotations

import argparse
import os
import secrets
import sys

import flet as ft
from loguru import logger

from libresvip.core.config import settings
from libresvip.core.constants import app_dir
from libresvip.mobile.app import main


def unraisablehook(unraisable: sys.UnraisableHookArgs) -> None:
    if unraisable.exc_type is not RuntimeError:
        logger.exception(unraisable.exc_value)
    del unraisable


sys.unraisablehook = unraisablehook


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--host", type=str, default="127.0.0.1")
    arg_parser.add_argument("--port", type=int, default=8080)
    arg_parser.add_argument("--web", action="store_true", help="Run in web mode")
    args, argv = arg_parser.parse_known_args()
    if args.web:
        import flet_web.fastapi
        import uvicorn
        from fastapi import FastAPI
        from fastapi.responses import FileResponse

        secrets_path = app_dir.user_config_path / "secrets.txt"
        if not secrets_path.exists():
            secrets_path.parent.mkdir(parents=True, exist_ok=True)
            secrets_path.write_text(secrets.token_urlsafe(32))
        os.environ["FLET_SECRET_KEY"] = secrets_path.read_text()
        flet_app = flet_web.fastapi.app(main, upload_dir=str(settings.save_folder))
        app = FastAPI()

        @app.get("/download/{filename}")
        def download(filename: str) -> FileResponse:
            return FileResponse(settings.save_folder / filename, filename=filename)

        app.mount("/", flet_app)

        uvicorn.run(app, host=args.host, port=args.port)
    else:
        ft.app(main, name="LibreSVIP")



import io
import pathlib
import sys

try:
    __import__("Cryptodome")
except ImportError:
    sys.modules["Cryptodome"] = __import__("Crypto")
from pyzipper import WZ_AES, ZIP_STORED, AESZipFile

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis

from .model import PocketSingerMetadata
from .options import InputOptions, OutputOptions
from .pocket_singer_generator import PocketSingerGenerator
from .pocket_singer_parser import PocketSingerParser

POCKET_SIGER_PASSWORD = b"a022ab39cb3b7b1de92ee441978c9e08"


class PocketSingerConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        with AESZipFile(
            io.BytesIO(path.read_bytes()), "r", compression=ZIP_STORED, encryption=WZ_AES
        ) as archive_file:
            archive_file.setpassword(POCKET_SIGER_PASSWORD)
            metadata = PocketSingerMetadata.model_validate_json(
                archive_file.read("config.json"),
                context={
                    "extract_audio": options.extract_audio,
                    "path": path,
                    "archive_file": archive_file,
                },
            )
            return PocketSingerParser(options, archive_file, path).parse_project(metadata)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        buffer = io.BytesIO()
        if len(project.song_tempo_list) != 1:
            project = reset_time_axis(project, options.tempo)
        with AESZipFile(buffer, "w", compression=ZIP_STORED, encryption=WZ_AES) as archive_file:
            archive_file.setpassword(POCKET_SIGER_PASSWORD)
            generator = PocketSingerGenerator(options)
            metadata = generator.generate_project(project)
            archive_file.writestr(
                "config.json",
                metadata.model_dump_json(exclude_none=True, by_alias=True).encode("utf-8"),
            )
            archive_file.writestr(metadata.ace_file_name, generator.buffer.getvalue())
            for audio_name, audio_path in generator.audio_paths.items():
                archive_file.writestr(audio_name, audio_path.read_bytes())
        path.write_bytes(buffer.getvalue())



import dataclasses
import functools
import math

import more_itertools
import portion

from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.utils.music_math import cosine_easing_in_out_interpolation

from .model import VocalSharpDefaultTrill, VocalSharpNoteTrack, VocalSharpTrill


def vspx_sine_vibrato_interpolation(
    seconds: float,
    vibrato_start: float,
    trill: VocalSharpTrill | VocalSharpDefaultTrill,
) -> float:
    return (
        math.sin(math.pi * (2 * (seconds - vibrato_start) * trill.frequency + trill.phase))
        * trill.amplitude
    )


def vspx_cosine_vibrato_coef_attack_interpolation(
    seconds: float, vibrato_start: float, por: float
) -> float:
    return 1 - math.cos(math.pi * (0.5 * (seconds - vibrato_start) / por))


def vspx_cosine_vibrato_coef_release_interpolation(
    seconds: float, vibrato_end: float, por: float
) -> float:
    return 1 - math.cos(math.pi * (0.5 * (vibrato_end - seconds) / por))


@dataclasses.dataclass
class BasePitchCurve:
    note_track: dataclasses.InitVar[VocalSharpNoteTrack]
    default_trill: dataclasses.InitVar[VocalSharpDefaultTrill | None]
    synchronizer: TimeSynchronizer
    key_interval_dict: PiecewiseIntervalDict = dataclasses.field(
        default_factory=PiecewiseIntervalDict
    )
    vibrato_value_interval_dict: PiecewiseIntervalDict = dataclasses.field(
        default_factory=PiecewiseIntervalDict
    )
    vibrato_coef_interval_dict: PiecewiseIntervalDict = dataclasses.field(
        default_factory=PiecewiseIntervalDict
    )

    def __post_init__(
        self,
        note_track: VocalSharpNoteTrack,
        default_trill: VocalSharpDefaultTrill | None,
    ) -> None:
        if not len(note_track.note):
            pass
        elif len(note_track.note) == 1:
            note = note_track.note[0]
            self.key_interval_dict[portion.closedopen(0, portion.inf)] = note.key_number
            if (trill := note.trill or default_trill) is not None:
                vibrato_start_secs = (
                    self.synchronizer.get_actual_secs_from_ticks(note.pos) + trill.pos
                )
                vibrato_end_secs = self.synchronizer.get_actual_secs_from_ticks(
                    note.pos + note.duration
                )
                if vibrato_end_secs > vibrato_start_secs:
                    self.set_vspx_vibrato_curve(
                        vibrato_start_secs,
                        vibrato_end_secs,
                        trill,
                        note_track.por,
                    )
        else:
            for is_first, is_last, (
                prev_note,
                next_note,
            ) in more_itertools.mark_ends(more_itertools.pairwise(note_track.note)):
                prev_start_secs = self.synchronizer.get_actual_secs_from_ticks(prev_note.pos)
                prev_end_secs = self.synchronizer.get_actual_secs_from_ticks(
                    prev_note.pos + prev_note.duration
                )
                next_start_secs = self.synchronizer.get_actual_secs_from_ticks(next_note.pos)
                next_end_secs = self.synchronizer.get_actual_secs_from_ticks(
                    next_note.pos + next_note.duration
                )
                if is_first:
                    self.key_interval_dict[portion.closedopen(0, prev_start_secs)] = (
                        prev_note.key_number
                    )
                middle_secs = (prev_end_secs + next_start_secs) / 2
                self.key_interval_dict[portion.closedopen(prev_start_secs, middle_secs)] = (
                    prev_note.key_number
                )
                self.key_interval_dict[
                    portion.closedopen(
                        middle_secs,
                        next_end_secs,
                    )
                ] = next_note.key_number
                if is_last:
                    self.key_interval_dict[portion.closedopen(next_start_secs, portion.inf)] = (
                        next_note.key_number
                    )
                    if (trill := next_note.trill or default_trill) is not None:
                        vibrato_start_secs = next_start_secs + trill.pos
                        vibrato_end_secs = next_end_secs
                        if vibrato_end_secs > vibrato_start_secs:
                            self.set_vspx_vibrato_curve(
                                vibrato_start_secs,
                                vibrato_end_secs,
                                trill,
                                note_track.por,
                            )
                if note_track.por > 0:
                    por_start = middle_secs - note_track.por
                    por_end = middle_secs + note_track.por
                    self.key_interval_dict[portion.closedopen(por_start, por_end)] = (
                        functools.partial(  # type: ignore[call-arg]
                            cosine_easing_in_out_interpolation,
                            start=(por_start, prev_note.key_number),
                            end=(por_end, next_note.key_number),
                        )
                    )
                if (trill := prev_note.trill or default_trill) is not None:
                    vibrato_start_secs = prev_start_secs + trill.pos
                    vibrato_end_secs = (prev_end_secs + next_start_secs) / 2
                    if vibrato_end_secs > vibrato_start_secs:
                        self.set_vspx_vibrato_curve(
                            vibrato_start_secs,
                            vibrato_end_secs,
                            trill,
                            note_track.por,
                        )

    def set_vspx_vibrato_curve(
        self,
        start: float,
        end: float,
        trill: VocalSharpTrill | VocalSharpDefaultTrill,
        por: float | None = None,
    ) -> None:
        self.vibrato_value_interval_dict[portion.closed(start, end)] = functools.partial(
            vspx_sine_vibrato_interpolation,
            vibrato_start=start,
            trill=trill,
        )
        if por is None or (end - start) < por * 2:
            middle = (start + end) / 2
            half = (end - start) / 2
            self.vibrato_coef_interval_dict[portion.closedopen(start, middle)] = functools.partial(
                vspx_cosine_vibrato_coef_attack_interpolation,
                vibrato_start=start,
                por=half,
            )
            self.vibrato_coef_interval_dict[portion.closed(middle, end)] = functools.partial(
                vspx_cosine_vibrato_coef_release_interpolation,
                vibrato_end=end,
                por=half,
            )
        elif por:
            self.vibrato_coef_interval_dict[portion.closedopen(start, start + por)] = (
                functools.partial(
                    vspx_cosine_vibrato_coef_attack_interpolation,
                    vibrato_start=start,
                    por=por,
                )
            )
            self.vibrato_coef_interval_dict[portion.closed(start + por, end - por)] = 1
            self.vibrato_coef_interval_dict[portion.openclosed(end - por, end)] = functools.partial(
                vspx_cosine_vibrato_coef_release_interpolation,
                vibrato_end=end,
                por=por,
            )
        else:
            self.vibrato_coef_interval_dict[portion.closed(start, end)] = 0

    def semitone_value_at(self, seconds: float) -> float | None:
        if (pitch_value := self.key_interval_dict.get(seconds)) is not None and (
            vibrato_value := self.vibrato_value_interval_dict.get(seconds)
        ) is not None:
            vibrato_value *= self.vibrato_coef_interval_dict[seconds]
            pitch_value += vibrato_value
        return pitch_value
